#!/usr/bin/env python3

import json
import argparse
import copy
import os
import subprocess

LINE_LIMIT = 58


def get_data(argsDict, resource_name):
    resource = []
    resource_sources = argsDict.get(resource_name, [])
    if not isinstance(resource_sources, list):
        resource_sources = [resource_sources]
    for resource_filename in resource_sources:
        if resource_filename.endswith(".json"):
            try:
                with open(
                        resource_filename, encoding="utf-8") as resource_file:
                    resource += json.load(resource_file)
            except FileNotFoundError:
                exit("Failed: could not find {}".format(resource_filename))
        else:
            print("Invalid filename {}".format(resource_filename))
    return resource


# stupid stinking Python 2 versus Python 3 syntax
def write_to_json(pathname, data):
    with open(pathname, "w", encoding="utf-8") as fp:
        try:
            json.dump(data, fp, ensure_ascii=False)
        except ValueError:
            fp.write(json.dumps(data))

    json_formatter = "./tools/format/json_formatter.cgi"
    if os.path.isfile(json_formatter):
        cmd = [json_formatter, pathname]
        subprocess.call(cmd)


def really_add_parts(xpoint, ypoint, xparts, new_parts):
    y_part = xparts.get(ypoint, {})
    if not y_part:
        return
    pretty_parts = []
    pretty_subparts = []
    temp_parts = ""
    for part in y_part["parts"]:
        temp_parts += '"{}"'.format(part)
        # debug - are we getting the line length test string correct?
        # print("{}: {}".format(len(temp_parts), temp_parts))
        if len(temp_parts) > LINE_LIMIT:
            pretty_parts.append(pretty_subparts)
            pretty_subparts = []
            temp_parts = '"{}"'.format(part)
        temp_parts += ", "
        pretty_subparts.append(part)
    pretty_parts.append(pretty_subparts)
    # debug - correct list of parts sublists?
    # print("{}".format(pretty_parts))
    for subpart_list in pretty_parts:
        if len(subpart_list) > 1:
            rev_part = {"x": xpoint, "y": ypoint, "parts": subpart_list}
        else:
            subpart = subpart_list[0]
            if isinstance(subpart, str):
                rev_part = {"x": xpoint, "y": ypoint, "part": subpart}
            else:
                rev_part = {"x": xpoint, "y": ypoint}
                for key, value in subpart.items():
                    if key == "x" or key == "y":
                        continue
                    rev_part[key] = value
        revised_parts.append(rev_part)


def add_parts(revised_parts, xpoint, last_center, new_parts):
    xparts = new_parts.get(xpoint, {})
    if not xparts:
        return 0
    min_y = 122
    max_y = -122
    for ypoint in xparts:
        min_y = min(min_y, ypoint)
        max_y = max(max_y, ypoint)

    if xpoint == 0:
        really_add_parts(xpoint, last_center, xparts, new_parts)
    for ypoint in range(min_y - 1, last_center, 1):
        really_add_parts(xpoint, ypoint, xparts, new_parts)
    if xpoint != 0:
        really_add_parts(xpoint, last_center, xparts, new_parts)
    for ypoint in range(last_center + 1, max_y + 1):
        really_add_parts(xpoint, ypoint, xparts, new_parts)
    return int((min_y + max_y) / 2)


args = argparse.ArgumentParser(
    description="Reformat vehicles using parts arrays.")
args.add_argument("vehicle_sources", action="store", nargs="+",
                  help="specify jsons file to convert to new format.")
argsDict = vars(args.parse_args())

for datafile in argsDict.get("vehicle_sources", []):
    min_x = 122
    max_x = -122
    vehicles = get_data(argsDict, "vehicle_sources")
    for old_vehicle in vehicles:
        #print(old_vehicle.get("id"))
        new_parts = {}
        for part in old_vehicle.get("parts", []):
            part_x = part.get("x", 0)
            part_y = part.get("y", 0)
            min_x = min(min_x, part_x)
            max_x = max(max_x, part_x)
            new_parts.setdefault(part_x, {})
            new_parts[part_x].setdefault(part_y, {"parts": []})
            part_list = []
            if part.get("parts"):
                for new_part in part.get("parts"):
                    new_parts[part_x][part_y]["parts"].append(new_part)
            else:
                if part.get("fuel") or part.get("ammo"):
                    new_part = copy.deepcopy(part)
                    del new_part["x"]
                    del new_part["y"]
                    new_parts[part_x][part_y]["parts"].append(new_part)
                else:
                    new_parts[part_x][part_y]["parts"].append(part.get("part"))

        # debug - did everything get copied over correctly?
        # print(json.dumps(new_parts, indent=2))
        revised_parts = []
        last_center = add_parts(revised_parts, 0, 0, new_parts)
        # print("last center {}".format(last_center))
        for xpoint in range(1, max_x + 1):
            last_center = add_parts(revised_parts, xpoint, last_center,
                                    new_parts)
            # print("x {}, last center {}".format(xpoint, last_center))
        for xpoint in range(-1, min_x - 1, -1):
            last_center = add_parts(revised_parts, xpoint, last_center,
                                    new_parts)

        old_vehicle["parts"] = revised_parts

    write_to_json(datafile, vehicles)



from ..helper import is_tag
from ..write_text import write_text


def parse_harvest(json, origin):
    if "message" in json:
        if not is_tag(json["message"]):
            write_text(json["message"], origin, comment="Harvest message")



from ..helper import get_singular_name
from ..write_text import write_text


def parse_technique(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Martial technique name")

    if "description" in json:
        write_text(json["description"], origin, c_format=False,
                   comment="Description of martial technique \"{}\""
                   .format(name))

    for msg in json.get("messages", []):
        write_text(msg, origin,
                   comment="Message of martial technique \"{}\"".format(name))

    if "condition_desc" in json:
        write_text(json["condition_desc"], origin, c_format=False,
                   comment="Condition description of martial technique \"{}\""
                   .format(name))



from ..write_text import write_text


def parse_sub_body_part(json, origin):
    write_text(json["name"], origin, comment="Name of sub body part")

    if "name_multiple" in json:
        write_text(json["name_multiple"], origin,
                   comment="Name of sub body part")



from ..helper import get_singular_name
from ..write_text import write_text


def parse_climbing_aid(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Name of a gun", plural=True)
    elif "id" in json:
        name = json["id"]

    down = None
    if "down" in json:
        down = json["down"]

    if down and "menu_text" in down:
        write_text(down["menu_text"], origin,
                   comment="Menu text for descent with \"{}\"".format(name))
    if down and "menu_cant" in down:
        write_text(down["menu_cant"], origin,
                   comment="Menu disables descent with \"{}\"".format(name))
    if down and "confirm_text" in down:
        write_text(down["confirm_text"], origin,
                   comment="Query: really descend with \"{}\"?".format(name))
    if down and "msg_before" in down:
        write_text(down["msg_before"], origin,
                   comment="Message before descent with \"{}\"".format(name))
    if down and "msg_after" in down:
        write_text(down["msg_after"], origin,
                   comment="Message after descent with \"{}\"".format(name))



#!/usr/bin/env python3
from spell_checker.spell_checker import spell_check
import re
import json
from optparse import OptionParser


parser = OptionParser()
parser.add_option("-i", "--input", dest="pot_diff",
                  help="Path to PO diff result in JSON format")

(options, args) = parser.parse_args()
if not options.pot_diff:
    parser.error("Input file not given")


diff = json.load(open(options.pot_diff, "r"))
errors = []
for message in diff["added"]:
    typos = set(spell_check(message))
    if len(typos) == 0:
        continue
    message = message.replace('\n', "\\n")
    for typo in typos:
        bold = "**" + typo + "**"
        message = re.sub(re.escape(typo), lambda _: bold, message)
    errors.append(message)
if errors:
    print("Spell checker encountered unrecognized words in the in-game text"
          " added in this pull request. See below for details.")
    print("<details>")
    print("<summary>Click to expand</summary>")
    print()
    for err in errors:
        print("*", err)
    print("</details>")
    print()
    print("This alert is automatically generated. You can simply disregard if"
          " this is inaccurate, or (optionally) you can also add the new words"
          " to `tools/spell_checker/dictionary.txt` so they will not trigger "
          "an alert next time.")
    print()
    print("<details>")
    print("<summary>Hints for adding a new word to the dictionary</summary>")
    print()
    print("* If the word is normally in all lowercase, such as the noun "
          "`word` or the verb `does`, add it in its lower-case form; if the "
          "word is a proper noun, such as the surname `George`, add it in its "
          "initial-caps form; if the word is an acronym or has special letter "
          "case, such as the acronym `CDDA` or the unit `mW`, add it by "
          "preserving the case of all the letters. A word in the dictionary "
          "will also match its initial-caps form (if the word is in all "
          "lowercase) and all-uppercase form, so a word should be added to "
          "the dictionary in its normal letter case even if used in a "
          "different letter case in a sentence.")
    print("* For a word to be added to the dictionary, it should either be a "
          "real, properly-spelled modern American English word, a foreign "
          "loan word (including romanized foreign names), or a foreign or "
          "made-up word that is used consistently and commonly enough in the "
          "game. Intentional misspelling (including eye dialect) of a word "
          "should not be added unless it has become a common terminology in "
          "the game, because while someone may have a legitimate use for it, "
          "another person may spell it that way accidentally.")
    print("</details>")



from ..write_text import write_text


def parse_recipe_group(json, origin):
    if "recipes" in json:
        for recipe in json["recipes"]:
            write_text(recipe["description"], origin,
                       comment="Description for recipe group")



from ..write_text import write_text


def parse_oter_vision(json, origin):
    if "levels" in json:
        for level in json["levels"]:
            if "name" in level:
                write_text(level["name"], origin,
                           comment="Overmap terrain vision level name")



#!/usr/bin/env python3

import json
import argparse
import os

# no one references these, but we may want them again some day
OBSOLETE_TOPICS = [
    "TALK_DENY_GUARD", "TALK_FRIEND_UNCOMFORTABLE", "TALK_USE_ITEM",
    "TALK_GIVE_ITEM"
]


args = argparse.ArgumentParser(
    description="Confirm that every talk topic in every response in a "
                "dialogue JSON is defined, and that every defined talk "
                "topic is referenced in at least one response.\n"
                "Also checks that every dialogue path reaches "
                "TALK_DONE and every topic is reachable from an NPC's"
                "starting topic.  Reports nothing on success.")
args.add_argument(
    "dialogue_json", nargs="+", action="store",
    help="specify json folder to validate.  The valdiator will walk the "
         "folder's tree and validate all JSON files in it.  Use "
         "'data/json/npcs/ to validate the dialogue in the vanilla game.")
argsDict = vars(args.parse_args())


def get_dialogue_from_json():
    dialogue = []

    for arg_path in argsDict.get("dialogue_json", []):
        if arg_path.endswith("/"):
            arg_path = arg_path[:-1]
        for subdir_path, dirnames, filenames in os.walk(arg_path):
            for filename in filenames:
                path = subdir_path + "/" + filename
                if path == "data/json/npcs/TALK_TEST.json":
                    continue
                if path.endswith(".json"):
                    with open(path, encoding="utf-8") as dialogue_file:
                        dialogue += json.load(dialogue_file)

    return dialogue


def add_topic_by_chat(topics, topic_id, topic_branches=None, this_ids=None):
    topics.setdefault(topic_id, {})
    if topic_branches is not None:
        topic_branches.setdefault(
            topic_id, {"responses": [], "ends": False, "parent": None})
    if this_ids is not None:
        this_ids.append(topic_id)
    topics[topic_id]["in_response"] = True


def add_topic_by_id(topics, topic_id, topic_branches=None, this_ids=None):
    if topic_id in OBSOLETE_TOPICS:
        return
    if topic_branches is not None:
        topic_branches.setdefault(
            topic_id, {"responses": [], "ends": False, "parent": None})
    if this_ids is not None:
        this_ids.append(topic_id)
    topics.setdefault(topic_id, {})
    topics[topic_id]["valid"] = True


def add_topic_by_response(topics, response, topic_branches=None,
                          this_ids=None):
    topic_id = response.get("topic")
    if not topic_id:
        return
    topics.setdefault(topic_id, {})
    topics[topic_id]["in_response"] = True
    if not topic_branches:
        return
    for parent_id in this_ids:
        if topic_id == "TALK_DONE":
            topic_branches[parent_id]["ends"] = True
        elif topic_id == "TALK_NONE":
            for previous_id in topic_branches:
                if parent_id in topic_branches[previous_id]["responses"]:
                    topic_branches[parent_id]["responses"].append(previous_id)
        else:
            topic_branches[parent_id]["responses"].append(topic_id)


def parse_response(topics, response, topic_branches=None, this_ids=None):
    if response.get("topic"):
        add_topic_by_response(topics, response, topic_branches, this_ids)
    else:
        success_r = response.get("success", {})
        add_topic_by_response(topics, success_r, topic_branches, this_ids)
        failure_r = response.get("failure", {})
        add_topic_by_response(topics, failure_r, topic_branches, this_ids)


def validate(dialogue):
    topics = {}
    topic_branches = {
        "TALK_TRAIN": {"responses": ["TALK_TRAIN_START"], "ends": False,
                       "parent": None},
        "TALK_HOW_MUCH_FURTHER": {"responses": ["TALK_DONE"], "ends": True,
                                  "parent": None},
        "TALK_SEDATED": {"responses": ["TALK_DONE"], "ends": True,
                         "parent": None},
    }
    # defined in src/npctalk.cpp
    defined_ids = [
        "TALK_NONE", "TALK_DONE", "TALK_TRAIN", "TALK_HOW_MUCH_FURTHER",
        "TALK_SEDATED"
    ]
    for topic_id in defined_ids:
        add_topic_by_id(topics, topic_id)

    # referenced in src/npctalk.cpp
    refered_ids = [
        "TALK_WAKE_UP", "TALK_RADIO", "TALK_MISSION_DESCRIBE_URGENT",
        "TALK_MISSION_DESCRIBE", "TALK_SHELTER", "TALK_SIZE_UP",
        "TALK_LOOK_AT", "TALK_OPINION", "TALK_SHOUT",
        "TALK_STRANGER_AGGRESSIVE", "TALK_LEADER", "TALK_TRAIN_START",
        "TALK_STOLE_ITEM", "TALK_SEDATED"
    ]
    for topic_id in refered_ids:
        add_topic_by_chat(topics, topic_id)

    # TALK_TRAIN_START is technically not a start id
    start_topics = [
        "TALK_RADIO", "TALK_LEADER", "TALK_FRIEND", "TALK_STOLE_ITEM",
        "TALK_MISSION_DESCRIBE_URGENT", "TALK_SEDATED", "TALK_WAKE_UP",
        "TALK_MUG", "TALK_STRANGER_AGGRESSIVE", "TALK_STRANGER_SCARED",
        "TALK_STRANGER_WARY", "TALK_STRANGER_FRIENDLY",
        "TALK_STRANGER_NEUTRAL", "TALK_SHELTER"
    ]
    for talk_topic in dialogue:
        if not isinstance(talk_topic, dict):
            continue
        if talk_topic.get("type") == "npc":
            chat = talk_topic.get("chat")
            if chat:
                start_topics.append(chat)
                add_topic_by_chat(topics, chat, topic_branches)
            continue
        if talk_topic.get("type") != "talk_topic":
            continue

        topic_id = talk_topic.get("id")
        this_ids = []
        if isinstance(topic_id, list):
            for loop_topic_id in topic_id:
                add_topic_by_id(topics, loop_topic_id, topic_branches,
                                this_ids)
        else:
            add_topic_by_id(topics, topic_id, topic_branches, this_ids)
        for response in talk_topic.get("responses", []):
            parse_response(topics, response, topic_branches, this_ids)
        repeated = talk_topic.get("repeat_responses", [])
        if isinstance(repeated, list):
            for repeated in talk_topic.get("repeat_responses", []):
                response = repeated.get("response", {})
                parse_response(topics, response, topic_branches, this_ids)
        else:
            response = repeated.get("response", {})
            parse_response(topics, response, topic_branches, this_ids)

    for topic_id, topic_record in topics.items():
        if not topic_record.get("valid", False):
            if topic_id in start_topics:
                print("talk topic {} referenced in an NPC chat but not "
                      "defined".format(topic_id))
            else:
                print("talk topic {} referenced in a response but not "
                      "defined".format(topic_id))
        if not topic_record.get("in_response", False):
            print("talk topic {} defined but not referenced in a "
                  "response".format(topic_id))
        if topic_id in OBSOLETE_TOPICS:
            print("talk topic {} referenced despite being listed as "
                  "obsolete.".format(topic_id))

    no_change = False
    passes = 0
    while not no_change and passes < len(topic_branches):
        no_change = True
        passes += 1
        for topic_id in topic_branches:
            branch_record = topic_branches[topic_id]
            if branch_record["ends"] or not topics[topic_id].get("valid",
                                                                 False):
                continue
            for response_id in branch_record["responses"]:
                if topic_branches.get(response_id, {}).get("ends", False):
                    # useful debug statement that I'm leaving here
                    #print("pass {} {} terminates by {}".format(
                    #    passes, topic_id, response_id))
                    branch_record["ends"] = True
                    no_change = False
                    break

    no_change = False
    passes = 0
    while not no_change and passes < len(topic_branches):
        no_change = True
        passes += 1
        for topic_id in topic_branches:
            # skip undefined topics
            if not topics[topic_id].get("valid", False):
                continue
            branch_record = topic_branches[topic_id]
            if branch_record["parent"] not in start_topics:
                if topic_id in start_topics:
                    branch_record["parent"] = topic_id
                    no_change = False
            if not branch_record["parent"]:
                continue
            for response_id in branch_record["responses"]:
                response_record = topic_branches.get(response_id, {})
                if response_record.get("parent", False) in start_topics:
                    continue
                response_record["parent"] = branch_record["parent"]
                no_change = False

    # print(json.dumps(topic_branches, indent=2))
    skip_topics = ["TALK_DONE", "TALK_SIZE_UP", "TALK_LOOK_AT", "TALK_OPINION",
                   "TALK_SHOUT"]

    for topic_id in topic_branches:
        if topic_id in skip_topics or not topics[topic_id].get("valid"):
            continue
        branch_record = topic_branches[topic_id]
        if not branch_record["ends"]:
            print("{} does not reach TALK_DONE".format(topic_id))
        if not branch_record["parent"] in start_topics:
            print("no path from a start topic to {}".format(topic_id))


validate(get_dialogue_from_json())



from ..helper import get_singular_name
from ..write_text import write_text


def parse_mapgen(json, origin):
    if "object" not in json:
        return

    om = ""
    if "om_terrain" in json:
        if type(json["om_terrain"]) is str:
            om = json["om_terrain"]
        elif type(json["om_terrain"]) is list:
            if len(json["om_terrain"]) == 0:
                om = ""
            elif type(json["om_terrain"][0]) is str:
                om = ", ".join(json["om_terrain"])
            elif type(json["om_terrain"][0]) is list:
                om = ", ".join(", ".join(i) for i in json["om_terrain"])

    parse_mapgen_object(json["object"], origin, om)


def parse_mapgen_object(json, origin, om):
    if type(json) is list:
        for j in json:
            parse_mapgen_object(j, origin, om)

    for key in ["place_specials", "place_signs"]:
        if key in json:
            for sign in json[key]:
                if "signage" in sign:
                    write_text(sign["signage"], origin,
                               comment="Signage placed on map {}".format(om))

    if "signs" in json:
        for sign in json["signs"]:
            if "signage" in json["signs"][sign]:
                write_text(json["signs"][sign]["signage"], origin,
                           comment="Signage placed on map {}".format(om))

    if "computers" in json:
        for key in json["computers"]:
            com = json["computers"][key]
            com_name = ""
            if "name" in com:
                com_name = get_singular_name(com["name"])
                write_text(
                    com["name"], origin,
                    comment="Computer name placed on map {}".format(om))
            for opt in com.get("options", []):
                if "name" in opt:
                    write_text(opt["name"], origin,
                               comment="Interactive menu name in computer "
                               "\"{}\" placed on map {}".format(com_name, om))
            if "access_denied" in com:
                write_text(com["access_denied"], origin,
                           comment="Access denied message on computer \"{}\""
                           " placed on map {}".format(com_name, om))

    if "place_computers" in json:
        for computer in json["place_computers"]:
            if "name" in computer:
                write_text(computer["name"], origin,
                           comment="Computer name placed on map {}".format(om))

    monsters = []
    if "place_monster" in json:
        monsters += json["place_monster"]
    if "monster" in json:
        monsters += [*json["monster"].values()]

    for m in monsters:
        if "name" in m:
            desc = ""
            if "monster" in m:
                desc = "\"{}\"".format(m["monster"])
            elif "group" in m:
                desc = "group \"{}\"".format(m["group"])
            write_text(m["name"], origin,
                       comment="Name of the monster {} placed on map {}"
                       .format(desc, om))

    if "place_graffiti" in json:
        for graffiti in json["place_graffiti"]:
            if "text" in graffiti:
                write_text(graffiti["text"], origin,
                           comment="Graffiti placed on map {}".format(om))



#!/usr/bin/env python3

# png_update.py
# Rename a png and update all references to it.

"""Rename a png file, its associated tile_entry.json, and update all
other tile_entry.json in the tileset dir to reflect the new name.
"""

import argparse
import json
import os
import subprocess


def write_to_json(pathname, data):
    with open(pathname, "w", encoding="utf-8") as fp:
        json.dump(data, fp, ensure_ascii=False)

    json_formatter = "./tools/format/json_formatter.cgi"
    if os.path.isfile(json_formatter):
        cmd = [json_formatter, pathname]
        subprocess.call(cmd)


def convert_index(old_index, old_name, new_name):
    changed = False
    new_indexes = []
    if isinstance(old_index, list):
        for pngname in old_index:
            if isinstance(pngname, dict):
                sprite_ids = pngname.get("sprite")
                new_sprites = []
                if isinstance(sprite_ids, list):
                    for sprite_id in sprite_ids:
                        if sprite_id == old_name:
                            new_sprites.append(new_name)
                            changed = True
                        else:
                            new_sprites.append(sprite_id)
                    if new_sprites:
                        pngname["sprite"] = new_sprites
                elif sprite_ids == old_name:
                    pngname["sprite"] = new_name
                    changed = True
                else:
                    pngname["sprite"] = sprite_ids
                new_indexes.append(pngname)
            elif pngname == old_name:
                new_indexes.append(new_name)
                changed = True
            else:
                new_indexes.append(pngname)
    elif old_index == old_name:
        new_indexes.append(new_name)
        changed = True
    else:
        new_indexes.append(old_index)

    if new_indexes and len(new_indexes) == 1:
        return new_indexes[0], changed
    return new_indexes, changed


def convert_tile_entry(tile_entry, old_name, new_name):
    if not isinstance(tile_entry, dict):
        return tile_entry, False

    changed = False
    old_fg_id = tile_entry.get("fg", [])
    if old_fg_id:
        tile_entry["fg"], fg_changed = convert_index(
            old_fg_id, old_name, new_name)
        changed = fg_changed

    old_bg_id = tile_entry.get("bg", [])
    if old_bg_id:
        tile_entry["bg"], bg_changed = convert_index(
            old_bg_id, old_name, new_name)
        changed |= bg_changed

    add_tile_entrys = tile_entry.get("additional_tiles", [])
    new_tile_entrys = []
    for add_tile_entry in add_tile_entrys:
        changed_tile_entry, add_changed = convert_tile_entry(
            add_tile_entry, old_name, new_name)
        new_tile_entrys.append(changed_tile_entry)
        changed |= add_changed
    if new_tile_entrys:
        tile_entry["additional_tiles"] = new_tile_entrys
    return tile_entry, changed


def convert_tile_entry_file(file_path, old_name, new_name):
    changed = False
    tile_data = []
    new_tile_data = []
    with open(file_path, "r", encoding="utf-8") as fp:
        tile_data = json.load(fp)
    if not isinstance(tile_data, list):
        tile_data = [tile_data]
    for tile_entry in tile_data:
        new_entry, entry_changed = convert_tile_entry(
            tile_entry, old_name, new_name)
        new_tile_data.append(new_entry)
        changed = entry_changed or changed
    if changed:
        if len(new_tile_data) == 1:
            new_tile_data = new_tile_data[0]
        write_to_json(file_path, new_tile_data)


if __name__ == '__main__':
    args = argparse.ArgumentParser(description=__doc__)
    args.add_argument("tileset_dir", action="store",
                      help="local name of the tileset directory under gfx/")
    args.add_argument("old_name", action="store",
                      help="old name of the png file")
    args.add_argument("new_name", action="store",
                      help="new name of the png file")
    argsDict = vars(args.parse_args())

    tileset_dirname = argsDict.get("tileset_dir", "")
    tmp_old_name = argsDict.get("old_name", "")
    tmp_new_name = argsDict.get("new_name", "")

    old_name = tmp_old_name
    new_name = tmp_new_name
    if tmp_old_name.endswith(".png"):
        old_name = tmp_old_name[:-4]
    if tmp_new_name.endswith(".png"):
        new_name = tmp_new_name[:-4]

    old_name_json = old_name + ".json"
    old_name_png = old_name + ".png"
    new_name_json = new_name + ".json"
    new_name_png = new_name + ".png"

    if not tileset_dirname.startswith("gfx/"):
        tileset_dirname = "gfx/" + tileset_dirname
    if tileset_dirname.endswith("/"):
        tileset_dirname = tileset_dirname[:-1]

    print("In " + tileset_dirname + ", renaming " +
          old_name + " to " + new_name)
    for png_dirname in os.listdir(tileset_dirname):
        if not png_dirname.startswith("pngs_"):
            continue
        png_path = tileset_dirname + "/" + png_dirname
        for subdir_fpath, dirnames, filenames in os.walk(png_path):
            for filename in filenames:
                old_path = subdir_fpath + "/" + filename
                if filename.endswith(".json"):
                    convert_tile_entry_file(old_path, old_name, new_name)
                if filename == old_name_png:
                    new_path = subdir_fpath + "/" + new_name_png
                    os.rename(old_path, new_path)
                elif filename == old_name_json:
                    new_path = subdir_fpath + "/" + new_name_json
                    os.rename(old_path, new_path)



#!/usr/bin/env python3

"""itemgroups reporter and checker"""

import argparse
import json
import os
import pprint
import sys
from pathlib import Path


TYPE_WHITELIST = [
    "AMMO", "ammobelt", "ARMOR", "BATTERY", "BIONIC_ITEM", "BOOK",
    "COMESTIBLE", "ENGINE", "GENERIC", "GUN", "GUNMOD", "holster",
    "MAGAZINE", "material", "musical_instrument", "mutagen", "mutagen_iv",
    "PET_ARMOR", "TOOL", "TOOL_ARMOR", "TOOLMOD", "WHEEL"
]
DEFAULT_CATEGORIES = [
    "AMMO", "ARMOR", "BIONIC_ITEM", "BOOK", "ENGINE", "GUN", "GUNMOD",
    "MAGAZINE", "PET_ARMOR", "TOOL", "TOOLMOD", "TOOL_ARMOR"
]
DATA_DIRECTORY = Path(os.path.dirname(__file__), "../..",
                      "data/json").resolve()
pp = pprint.PrettyPrinter(indent=2)


def _parse_arguments():
    parser = argparse.ArgumentParser(
        description="This is an itemgroup reporter and linter.  "
                    "This object menaces with subtle bugs and "
                    "recursive apocalypses.",
        epilog="Do not blindly trust the output of this program")
    parser.add_argument("-o", "--orphans", action="store_true",
                        help="Display orphaned items")
    parser.add_argument("-m", "--map", action="store_true",
                        help="Display item to item_group mapping")
    cathelp = "Limit output to these categories (default: '%s').  " \
              "'CATEGORIES' is a comma-separated list, like " \
              "'GUN, MAGAZINE'. Do not forget to quote the argument " \
              "if necessary." % (", ".join(DEFAULT_CATEGORIES))
    parser.add_argument("-c", "--categories", help=cathelp)
    return parser.parse_args()


def load_json_data(data_directory):
    """Load the JSON data.
       Returns:
        (itemgroups, items)"""
    entries = []
    ignore_filenames = ["obsolete.json", "migration.json"]
    for filename in data_directory.glob("**/*.json"):
        if filename.name in ignore_filenames:
            continue
        try:
            with open(filename, encoding="utf-8") as json_fp:
                json_data = json.load(json_fp)
            for data in json_data:
                if isinstance(data, str):
                    continue
                data["original_filename"] = \
                    filename.relative_to(data_directory)
                entries.append(data)
        except AttributeError as err:
            print("ERROR: Failed to read data from '%s': %s" %
                  (filename, err), file=sys.stderr)
            print(data)
            sys.exit(1)
    return entries


def _add_to_group(item_id, group, group_id):
    """Helper function to add an item id to the group map"""
    if item_id not in group:
        group[item_id] = {group_id: 1}
    elif group_id not in group[item_id]:
        group[item_id][group_id] = 1
    else:
        group[item_id][group_id] += 1


def _recurse_through_igroups(igroup, item_group, item_to_group, group_id=None):
    """Recurses through the item_group hierarchy and tries
       to resolve the various item groups, and fill out
       the item_to_group map as it goes"""
    if isinstance(igroup, dict):
        group_id = igroup.get("id")
        entries = igroup.get("entries", [])
        if not entries:
            entries = igroup.get("items", [])
        if not entries:
            groups = igroup.get("groups", [])
            if groups:
                for group in groups:
                    if isinstance(group, list):
                        groupname = group[0]
                    else:
                        groupname = group
                    if groupname in item_group:
                        _recurse_through_igroups(item_group[groupname],
                                                 item_group,
                                                 item_to_group,
                                                 group_id=group_id)
                    else:
                        print("ERROR: '%s' does not have 'items', or "
                              "'entries', or usable 'groups'" %
                              group_id, file=sys.stderr)
                        pp.pprint(igroup)
                        sys.exit(1)
    else:
        entries = igroup

    for entry in entries:
        if isinstance(entry, list):
            itemname = entry[0]
            if isinstance(itemname, str):
                _add_to_group(itemname, item_to_group, group_id)
            continue
        if "item" in entry:
            _add_to_group(entry["item"], item_to_group, group_id)
            continue
        if "group" in entry:
            if entry["group"] in item_group:
                _recurse_through_igroups(item_group[entry["group"]],
                                         item_group, item_to_group)
            continue
        if "distribution" in entry:
            _recurse_through_igroups(entry["distribution"], item_group,
                                     item_to_group, group_id)
            continue
        if "collection" in entry:
            _recurse_through_igroups(entry["collection"],
                                     item_group, item_to_group, group_id)
            continue


def get_item_data(entries, categories=None, ignore=None):
    """Scans the raw data structure from JSON and constructs
       an item to group map, a dict of orphans, a dict of items,
       an a list of potential problems"""
    ignore_items = ["battery_test"]
    TYPE_WHITELIST.append("item_group")
    if ignore:
        ignore_items = ignore
    item_categories = DEFAULT_CATEGORIES
    if categories:
        item_categories = categories
    item_group = {}
    item_entry = {}
    item_orphan = {}
    item_to_group = {}
    problems = []
    for entry in entries:
        copy_from = entry.get("copy-from", "")
        if copy_from == "fake_item":
            continue
        entry_type = entry.get("type")
        path = entry.pop("original_filename")
        if not entry_type or entry_type not in TYPE_WHITELIST:
            continue
        if entry.get("id") in ignore_items:
            continue
        if entry_type == "item_group":
            igroup_id = entry.get("id")
            if not igroup_id:
                problems.append({"type": "missing id", "path": path,
                                 "entry": entry})
            item_group[igroup_id] = entry
            continue

        # if it's not an item_group, it's probably an item
        item_id = entry.get("id")
        item_entry[item_id] = entry
    for igroup in item_group:
        _recurse_through_igroups(item_group[igroup], item_group, item_to_group)
    for item in item_entry:
        itemtype = item_entry[item].get("type")
        if itemtype not in item_categories:
            continue
        if item not in item_to_group:
            item_orphan[item] = item_entry[item]
    return (item_to_group, item_orphan, item_entry, problems)


if __name__ == "__main__":
    args = _parse_arguments()
    data_entries = load_json_data(DATA_DIRECTORY)
    item_categories = DEFAULT_CATEGORIES
    if args.categories:
        item_categories = [c.strip(' ') for c in
                           args.categories.split(",")]
    (itemgroup, orphan, items, problems) = \
        get_item_data(data_entries, item_categories)
    if problems:
        for problem in problems:
            print(problem, file=sys.stderr)
    if args.orphans:
        print("Orphans (items not found in any item groups):")
        print("=============================================")
        for oid in orphan:
            name = orphan[oid].get("name")
            if not name:
                name = "*no name entry*"
            if isinstance(name, dict):
                if "str" in name:
                    name = name["str"]
                elif "str_sp" in name:
                    name = name["str_sp"]
                elif "str_pl" in name:
                    name = name["str_pl"]
            else:
                print("WARNING: the 'name' property of '%s' is not a dict" %
                      (oid), file=sys.stderr)
                print(orphan[oid], file=sys.stderr)
            print("%s ('%s')" % (oid, name))
    if args.map:
        print("item to itemgroup mapping:")
        print("==========================")
        for item in itemgroup:
            if item not in items:
                continue
            itemtype = items[item].get("type", "")
            if itemtype not in item_categories:
                continue
            groups = list(itemgroup[item].keys())
            print("%s: %s" % (item, ", ".join(groups)))
    if not args.orphans and not args.map:
        print("%d items in %d item groups in categories:\n%s\n"
              "Default categories:\n%s" %
              (len(items), len(itemgroup),
               ", ".join(TYPE_WHITELIST),
               ", ".join(DEFAULT_CATEGORIES)))
        print("Use -h or --help for more comprehensive information")



#!/usr/bin/env python3

import sys
import argparse
import json


def translate(value, value_min_range, value_max_range, min_range, max_range):
    value_span = value_max_range - value_min_range
    span = max_range - min_range
    scaled = float(value - value_min_range) / float(value_span)
    return min_range + (scaled * span)


def rgb_as_int(rgb1, rgb2):
    n_rgb1 = tuple([translate(c, 0, 255, 0, 1) for c in rgb1])
    n_rgb2 = tuple([translate(c, 0, 255, 0, 1) for c in rgb2])
    return rgb(n_rgb1, n_rgb2)


# https://www.w3.org/TR/WCAG21/#dfn-contrast-ratio
def rgb(rgb1, rgb2):
    for r, g, b in (rgb1, rgb2):
        if not 0.0 <= r <= 1.0:
            raise ValueError("r is out of valid range (0.0 - 1.0)")
        if not 0.0 <= g <= 1.0:
            raise ValueError("g is out of valid range (0.0 - 1.0)")
        if not 0.0 <= b <= 1.0:
            raise ValueError("b is out of valid range (0.0 - 1.0)")

    l1 = _relative_luminance(*rgb1)
    l2 = _relative_luminance(*rgb2)

    if l1 > l2:
        contrast = (l1 + 0.05) / (l2 + 0.05)
    else:
        contrast = (l2 + 0.05) / (l1 + 0.05)

    return round(contrast, 3)


# https://www.w3.org/TR/WCAG21/#dfn-relative-luminance
def _relative_luminance(r, g, b):
    r = _linearize(r)
    g = _linearize(g)
    b = _linearize(b)

    return 0.2126 * r + 0.7152 * g + 0.0722 * b


def _linearize(v):
    if v <= 0.03928:
        return v / 12.92
    else:
        return ((v + 0.055) / 1.055) ** 2.4


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-j",
        "--colors",
        dest="base_colors",
        type=str,
        help="path to base colors json",
        default="",
    )
    parser.add_argument(
        "-p",
        "--print",
        dest="print",
        type=str,
        help="""print out color pairs, you'll need terminal with 24-bit color
        support enabled; enabled by default""",
        default="yes",
        choices=["yes", "no"],
    )
    parser.add_argument(
        "-s",
        "--sort",
        dest="sort",
        action="store_true",
        help="""should the list of color pairs be ordered in descending
        contrast?""",
        default=False,
    )
    parser.add_argument(
        "-f",
        "--filter",
        dest="filter",
        type=str,
        help="Color to filter on"
    )
    args = parser.parse_args(args=None if sys.argv[1:] else ["--help"])

    with open(args.base_colors, "rt") as f:
        bc_json = json.load(f)[0]

    colors = []
    column_width = 0
    for record in bc_json:
        if record != "type":
            column_width = max(column_width, len(record) + 2)
            pair = {
                "name": record,
                "rgb": (
                    bc_json[record][0],
                    bc_json[record][1],
                    bc_json[record][2]
                ),
            }
            colors.append(pair)
    ls = []
    for c1 in colors:
        if args.filter and c1["name"] != args.filter:
            continue
        for c2 in colors:
            if c1["name"] != c2["name"]:
                ls.append((rgb_as_int(c1["rgb"], c2["rgb"]), c1, c2))
    if args.sort:
        ls = sorted(ls, key=lambda x: x[0], reverse=True)
    for v, c1, c2 in ls:
        if v >= 7.0:
            v = "AAA"
        elif v >= 4.5:
            v = "AA"
        s_color_1 = f"{c1['name']:>{column_width}s}"
        s_color_2 = f"{c2['name']:>{column_width}s}"
        s = f"{s_color_1} {s_color_2} {v:>7}"
        if args.print == "yes":
            s += (f" \x1b[48;2;{c1['rgb'][0]};{c1['rgb'][1]};{c1['rgb'][2]}m"
                  f"\x1b[38;2;{c2['rgb'][0]};{c2['rgb'][1]};{c2['rgb'][2]}m"
                  f" {c2['rgb'][0]:3} {c2['rgb'][1]:3} {c2['rgb'][2]:3}"
                  " \x1b[0m")
        print(s)



#!/usr/bin/env python3
"""Count the number of times a specific key occurs.

Example usages:

    # List all keys, human friendly.
    %(prog)s --human

    # List keys on JSON objects of type "bionic", output in JSON.
    %(prog)s type=bionic

"""

import sys
import json
import argparse
from util import import_data, key_counter, ui_counts_to_columns, WhereAction

parser = argparse.ArgumentParser(
    description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)

parser.add_argument(
    "--fnmatch",
    default="*.json",
    help="override with glob expression to select a smaller fileset.")
parser.add_argument(
    "-H", "--human",
    action="store_true",
    help="if set, makes output human readable. default is JSON output.")
parser.add_argument(
    "-L", "--list",
    action="store_true",
    help="print only a sorted list of keys in JSON format,"
         " or newline-separated plain text list if --human is also set.")
parser.add_argument(
    "where",
    action=WhereAction, nargs='*', type=str,
    help="where exclusions of the form 'where_key=where_val', no quotes.")


if __name__ == "__main__":
    args = parser.parse_args()

    json_data, load_errors = import_data(json_fmatch=args.fnmatch)
    if load_errors:
        # If we start getting unexpected JSON or other things, might need to
        # revisit quitting on load_errors
        print("Error loading JSON data.")
        for e in load_errors:
            print(e)
        sys.exit(1)
    elif not json_data:
        print("No data loaded.")
        sys.exit(1)

    stats, num_matches = key_counter(json_data, args.where)

    if not stats:
        print("Nothing found.")
        sys.exit(1)

    if args.human and args.list:
        print("\n".join(sorted(stats.keys())))
    elif args.human:
        title = "Count of keys"
        print("\n\n%s" % title)
        print("(Data from %s out of %s blobs)" % (num_matches, len(json_data)))
        print("-" * len(title))
        ui_counts_to_columns(stats)
    elif args.list:
        print(json.dumps(sorted(stats.keys())))
    else:
        print(json.dumps(stats))



from ..helper import get_singular_name
from ..write_text import write_text


def parse_fault(json, origin):
    name = get_singular_name(json["name"])
    write_text(json["name"], origin, comment="Name of a fault")
    write_text(json["description"], origin,
               comment="Description of fault \"{}\"".format(name))
    if "item_prefix" in json:
        write_text(json["item_prefix"], origin,
                   comment="Prefix for item's name \"{}\"".format(name))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""""
Requires: python 3.6+; Windows 7+
Code format: PEP-8; line breaks at 120; Black default formatting.

Limits a target process to a given amount of memory.
If the process use more memory than the given limit, the process allocations
will start to fail.

```
$ python .\\windows_limit_memory.py --help
usage: windows_limit_memory.py [-h]
                               [-l {NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL}]
                               [-m MEMORY]
                               {process,pid} ...

C:DDA Memory Limit test script.

positional arguments:
  {process,pid}         help for sub-commands
    process             Start a new process.
    pid                 Enforce limit on a given process pid.

optional arguments:
  -h, --help            show this help message and exit
  -l,--log-level {NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL}
                        Set the logging level.
  -m MEMORY, --memory MEMORY
                        Maximum process memory size in MiB.
```

Examples:
    ; limiting memory of process with PID 22764 to 50 MiB.
    $ python windows_limit_memory.py -m 50 pid 22764

    ; starting process "cataclysm-tiles.exe" and limiting its memory to 1GiB.
    $ python windows_limit_memory.py -m 1024 process \
            z:\\CDDA\\Cataclysm-tiles.exe
"""

import argparse
import ctypes
import logging
import pathlib
import platform
import sys
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)

#
# Windows basic types
#
LPCWSTR = ctypes.c_wchar_p
BOOL = ctypes.c_long
HANDLE = ctypes.c_void_p
WORD = ctypes.c_uint16
DWORD = ctypes.c_uint32
if ctypes.sizeof(ctypes.c_void_p) == 8:
    ULONG_PTR = ctypes.c_ulonglong
else:
    ULONG_PTR = ctypes.c_ulong
SIZE_T = ULONG_PTR
PVOID = ctypes.c_void_p
LPVOID = PVOID
LPDWORD = ctypes.POINTER(DWORD)
PULONG_PTR = ctypes.POINTER(ULONG_PTR)

#
# Windows defines
#
CREATE_SUSPENDED = 0x4
CREATE_BREAKAWAY_FROM_JOB = 0x01000000

# -1 on 32 or 64-bit.
INVALID_HANDLE_VALUE = 2 ** (ctypes.sizeof(ctypes.c_void_p) * 8) - 1
JobObjectAssociateCompletionPortInformation = 7
JobObjectExtendedLimitInformation = 9
JOB_OBJECT_LIMIT_PROCESS_MEMORY = 0x100
PROCESS_SET_QUOTA = 0x100
PROCESS_TERMINATE = 0x1
INFINITE = 0xFFFFFFFF

# Completion Port Messages for job objects
JOB_OBJECT_MSG_END_OF_JOB_TIME = 1
JOB_OBJECT_MSG_END_OF_PROCESS_TIME = 2
JOB_OBJECT_MSG_ACTIVE_PROCESS_LIMIT = 3
JOB_OBJECT_MSG_ACTIVE_PROCESS_ZERO = 4
JOB_OBJECT_MSG_NEW_PROCESS = 6
JOB_OBJECT_MSG_EXIT_PROCESS = 7
JOB_OBJECT_MSG_ABNORMAL_EXIT_PROCESS = 8
JOB_OBJECT_MSG_PROCESS_MEMORY_LIMIT = 9
JOB_OBJECT_MSG_JOB_MEMORY_LIMIT = 10
JOB_OBJECT_MSG_NOTIFICATION_LIMIT = 11
JOB_OBJECT_MSG_JOB_CYCLE_TIME_LIMIT = 12
JOB_OBJECT_MSG_SILO_TERMINATED = 13

#
# Windows Structures
#


class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = [("nLength", DWORD), ("lpSecurityDescriptor", LPVOID),
                ("bInheritHandle", BOOL)]


class STARTUPINFO(ctypes.Structure):
    _fields_ = [
        ("cb", DWORD),
        ("lpReserved", LPCWSTR),
        ("lpDesktop", LPCWSTR),
        ("lpTitle", LPCWSTR),
        ("dwX", DWORD),
        ("dwY", DWORD),
        ("dwXSize", DWORD),
        ("dwYSize", DWORD),
        ("dwXCountChars", DWORD),
        ("dwYCountChars", DWORD),
        ("dwFillAttribute", DWORD),
        ("dwFlags", DWORD),
        ("wShowWindow", WORD),
        ("cbReserved2", WORD),
        ("lpReserved2", LPVOID),
        ("hStdInput", HANDLE),
        ("hStdOutput", HANDLE),
        ("hStdError", HANDLE),
    ]


class PROCESS_INFORMATION(ctypes.Structure):
    _fields_ = [("hProcess", HANDLE), ("hThread", HANDLE),
                ("dwProcessId", DWORD), ("dwThreadId", DWORD)]


class IO_COUNTERS(ctypes.Structure):
    _fields_ = [
        ('ReadOperationCount', ctypes.c_ulonglong),
        ('WriteOperationCount', ctypes.c_ulonglong),
        ('OtherOperationCount', ctypes.c_ulonglong),
        ('ReadTransferCount', ctypes.c_ulonglong),
        ('WriteTransferCount', ctypes.c_ulonglong),
        ('OtherTransferCount', ctypes.c_ulonglong),
    ]


class JOBOBJECT_BASIC_LIMIT_INFORMATION(ctypes.Structure):
    _fields_ = [
        ('PerProcessUserTimeLimit', ctypes.c_int64),
        ('PerJobUserTimeLimit', ctypes.c_int64),
        ('LimitFlags', ctypes.c_uint32),
        ('MinimumWorkingSetSize', ctypes.c_ulonglong),
        ('MaximumWorkingSetSize', ctypes.c_ulonglong),
        ('ActiveProcessLimit', ctypes.c_uint32),
        ('Affinity', ctypes.c_void_p),
        ('PriorityClass', ctypes.c_uint32),
        ('SchedulingClass', ctypes.c_uint32),
    ]


class JOBOBJECT_EXTENDED_LIMIT_INFORMATION(ctypes.Structure):
    _fields_ = [
        ('BasicLimitInformation', JOBOBJECT_BASIC_LIMIT_INFORMATION),
        ('IoInfo', IO_COUNTERS),
        ('ProcessMemoryLimit', ctypes.c_ulonglong),
        ('JobMemoryLimit', ctypes.c_ulonglong),
        ('PeakProcessMemoryUsed', ctypes.c_ulonglong),
        ('PeakJobMemoryUsed', ctypes.c_ulonglong),
    ]


class JOBOBJECT_ASSOCIATE_COMPLETION_PORT(ctypes.Structure):
    _fields_ = [('CompletionKey', LPVOID), ('CompletionPort', HANDLE)]


class OVERLAPPED(ctypes.Structure):
    _fields_ = [("Internal", ULONG_PTR), ("InternalHigh", ULONG_PTR),
                ("Pointer", LPVOID), ("hEvent", HANDLE)]


LPSECURITY_ATTRIBUTES = ctypes.POINTER(SECURITY_ATTRIBUTES)
LPSTARTUPINFO = ctypes.POINTER(STARTUPINFO)
LPPROCESS_INFORMATION = ctypes.POINTER(PROCESS_INFORMATION)
LPOVERLAPPED = ctypes.POINTER(OVERLAPPED)


class Kernel32Wrapper:
    """A class used to encapsulate kernel32 library functions.
    """

    def __init__(self) -> None:
        """Initialization.
        """
        # get kernel32 instance
        self._kernel32: ctypes.WinDLL = ctypes.WinDLL(
            "kernel32", use_last_error=True)

        #
        # fetch functions
        #

        self._create_job_object = self._kernel32.CreateJobObjectW
        self._create_job_object.argtypes = (LPSECURITY_ATTRIBUTES, LPCWSTR)
        self._create_job_object.restype = HANDLE

        self._query_information_job_object = \
            self._kernel32.QueryInformationJobObject
        self._query_information_job_object.argtypes = (
            HANDLE, ctypes.c_uint32, LPVOID, DWORD, ctypes.POINTER(DWORD))
        self._query_information_job_object.restype = BOOL

        self._set_information_job_object = \
            self._kernel32.SetInformationJobObject
        self._set_information_job_object.argtypes = (
            HANDLE, ctypes.c_uint32, LPVOID, DWORD)
        self._set_information_job_object.restype = BOOL

        self._assign_process_to_job_object = \
            self._kernel32.AssignProcessToJobObject
        self._assign_process_to_job_object.argtypes = (HANDLE, HANDLE)
        self._assign_process_to_job_object.restype = BOOL

        self._create_process = self._kernel32.CreateProcessW
        self._create_process.argtypes = (
            LPCWSTR,
            LPCWSTR,
            LPSECURITY_ATTRIBUTES,
            LPSECURITY_ATTRIBUTES,
            BOOL,
            DWORD,
            LPVOID,
            LPCWSTR,
            LPSTARTUPINFO,
            LPPROCESS_INFORMATION,
        )
        self._create_process.restype = BOOL

        self._open_process = self._kernel32.OpenProcess
        self._open_process.argtypes = (DWORD, BOOL, HANDLE)
        self._open_process.restype = HANDLE

        self._create_io_completion_port = self._kernel32.CreateIoCompletionPort
        self._create_io_completion_port.argtypes = (
            HANDLE, HANDLE, ULONG_PTR, DWORD)
        self._create_io_completion_port.restype = HANDLE

        self._get_queued_completion_status = \
            self._kernel32.GetQueuedCompletionStatus
        self._get_queued_completion_status.argtypes = (
            HANDLE, LPDWORD, PULONG_PTR, ctypes.POINTER(LPOVERLAPPED),
            DWORD)
        self._get_queued_completion_status.restype = BOOL

        self._resume_thread = self._kernel32.ResumeThread
        self._resume_thread.argtypes = (HANDLE,)
        self._resume_thread.restype = DWORD

        self._terminate_process = self._kernel32.TerminateProcess
        self._terminate_process.argtypes = (HANDLE, ctypes.c_uint32)
        self._terminate_process.restype = BOOL

        self._close_handle = self._kernel32.CloseHandle
        self._close_handle.argtypes = (HANDLE,)
        self._close_handle.restype = BOOL

        self._function_map = {
            "AssignProcessToJobObject": self._assign_process_to_job_object,
            "CloseHandle": self._close_handle,
            "CreateIoCompletionPort": self._create_io_completion_port,
            "CreateJobObject": self._create_job_object,
            "CreateProcess": self._create_process,
            "GetQueuedCompletionStatus": self._get_queued_completion_status,
            "OpenProcess": self._open_process,
            "QueryInformationJobObject": self._query_information_job_object,
            "ResumeThread": self._resume_thread,
            "TerminateProcess": self._terminate_process,
            "SetInformationJobObject": self._set_information_job_object,
        }

    def __getattr__(self, item: str):
        """Get an attribute from the class.

        Args:
            item: The attribute to get.

        Raises:
            AttributeError: the given name is not a known function name.

        Returns:
            If the attribute is a function name, returns the function pointer,
            otherwise raise an exception.
        """
        func = self._function_map.get(item)
        if func is None:
            raise AttributeError(item)
        return func

    @staticmethod
    def create_buffer(obj: Any, max_buffer_len: Optional[int] = None) -> str:
        """Creates a ctypes unicode buffer given an object convertible to
            string.

        Args:
            obj: The object from which to create a buffer. Must be convertible
                to `str`.
            max_buffer_len: The maximum buffer length for the buffer. If `None`
                is supplied, default to the length of
                the given object string.

        Returns:
            A unicode buffer of object converted to string.
        """
        str_obj = obj if isinstance(obj, str) else str(obj)

        if max_buffer_len is None:
            max_len_value = len(str_obj) + 1
        else:
            max_len_value = max_buffer_len
        max_len = max(max_len_value, len(str_obj))

        return ctypes.create_unicode_buffer(str_obj, max_len)


class ProcessLimiter:
    """A class used to limit a process memory using Windows Jobs.
    """

    def __init__(self) -> None:
        """Initialization.
        """
        logger.debug("instantiating kernel32 wrapper.")
        self._kernel32: Kernel32Wrapper = Kernel32Wrapper()
        self._handle_process: Optional[HANDLE] = None
        self._handle_thread: Optional[HANDLE] = None
        self._handle_job: Optional[HANDLE] = None
        self._handle_io_port: Optional[HANDLE] = None
        try:
            self._handle_io_port: HANDLE = self._create_io_completion_port()
        except ctypes.WinError:
            pass

    def __enter__(self) -> "ProcessLimiter":
        """Context manager entry.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Context manager exit.

        Args:
            exc_type: exception type.
            exc_val: exception value.
            exc_tb: Exception trackeback.
        """
        if self._handle_process:
            self._kernel32.CloseHandle(self._handle_process)
            self._handle_process = None
        if self._handle_thread:
            self._kernel32.CloseHandle(self._handle_thread)
            self._handle_thread = None
        if self._handle_io_port:
            self._kernel32.CloseHandle(self._handle_io_port)
            self._handle_io_port = None
        if self._handle_job:
            self._kernel32.CloseHandle(self._handle_job)
            self._handle_job = None

    @property
    def has_io_port(self) -> bool:
        """Get whether the current class instance holds a Windows I/O port.

        Returns:
            `True` if the class hold an I/O port, `False` otherwise.
        """
        return self._handle_io_port is not None

    @property
    def is_started_process(self) -> bool:
        """Get whether the process object was started by the class or not.

        Returns:
            True if the process was started by this class instance, False
            otherwise.
        """
        return (self._handle_process is not None and
                self._handle_thread is not None)

    def _create_io_completion_port(self) -> HANDLE:
        """[Internal] Create an I/O completion port.

        Notes:
            The I/O port is used later in the `wait_for_job()` function.

        Raises:
            ctypes.WinError: An error occurred while creating the I/O port.

        Returns:
            A `HANDLE` to the I/O Port.
        """
        logger.info("Creating IO Port")
        handle_io_port = self._kernel32.CreateIoCompletionPort(
            INVALID_HANDLE_VALUE, None, 0, 1)
        if not handle_io_port:
            raise ctypes.WinError(ctypes.get_last_error())
        logger.debug(f"IO Port: {handle_io_port:#x}")
        return handle_io_port

    def _query_information_job_object(self, structure: ctypes.Structure,
                                      query_type: int) -> ctypes.Structure:
        """[Internal] Retrieves limit and job state information from the job
            object.

        Args:
            structure: The limit or job state information.
            query_type: The information class for the limits to be queried.

        Raises:
            ctypes.WinError: An error occurred while getting the state or limit
                for the job object.

        Returns:
            Returns the given structure filled with the job limit or state
            information.
        """
        # query its default properties
        return_length = DWORD(0)
        logger.debug("Querying job object.")
        ret_val = self._kernel32.QueryInformationJobObject(
            self._handle_job, query_type, ctypes.byref(structure),
            ctypes.sizeof(structure), ctypes.byref(return_length)
        )
        if ret_val == 0 or return_length.value != ctypes.sizeof(structure):
            raise ctypes.WinError(ctypes.get_last_error())
        return structure

    def _resume_main_thread(self) -> None:
        """[Internal] Resume the main thread of a created process.

        Raises:
            ValueError: the function was called but there is no main thread
                handle.
            ctypes.WinError: There was an error while trying to resume the main
                thread of the process. Note that this is a critical condition
                resulting in a zombie process. The code will try to kill the
                zombie process if it happens, without any guaranty of success.
        """
        if not self._handle_thread:
            raise ValueError("Thread handle is NULL.")
        # resume the main thread and let the process run.
        logger.debug("Resuming main thread.")
        ret_val = self._kernel32.ResumeThread(self._handle_thread)
        if ret_val == -1:
            # oops, we now have a zombie process... we'll try to kill it
            # nonetheless.
            logger.error(
                "Error: ResumeThread failed but the process is started!")
            # try to kill it
            if 0 != self._kernel32.TerminateProcess(self._handle_process,
                                                    0xDEADBEEF):
                logger.info("Successfully killed the zombie process.")
            else:
                logger.warning("The zombie process is sitll alive. Try to "
                               "kill it manually.")
            # we tried to kill the process, now raise.
            raise ctypes.WinError(ctypes.get_last_error())

    def _set_information_job_object(self, structure: ctypes.Structure,
                                    set_type: int) -> None:
        """[Internal] Set limit and job state information for the job object.

        Args:
            structure: The limit or job state information.
            set_type: The information class for the limits to be queried.

        Raises:
            ctypes.WinError: An error occurred while setting the state or limit
                for the job object.
        """
        ret_val = self._kernel32.SetInformationJobObject(
            self._handle_job, set_type, ctypes.byref(structure),
            ctypes.sizeof(structure)
        )
        if ret_val == 0:
            raise ctypes.WinError(ctypes.get_last_error())

    def assign_process_to_job(self) -> None:
        """Assign a process to a job.

        Raises:
            ValueError: There's no process or job to associate with.
            ctypes.WinError: There was an error while associating the process
                with the job.
        """
        if not self._handle_process:
            raise ValueError("There's no process to associate with the job.")
        if not self._handle_job:
            raise ValueError("There's no job.")
        logger.info("Assigning process to job.")
        ret_val = self._kernel32.AssignProcessToJobObject(
            self._handle_job, self._handle_process)
        if ret_val == 0:
            raise ctypes.WinError(ctypes.get_last_error())

    def create_job(self, job_name: Optional[str] = None) -> None:
        """Create a job.

        Args:
            job_name: The optional job name; can be `None`.

        Notes:
            This function will try to associate the job with an I/O completion
            port, which is used later in the `wait_for_job` function.

        Raises:
            ctypes.WinError: There was an error while creating the job, or, if
                a completion port exists, an error
                occurred while associating the completion port to the job.
        """
        logger.info("Creating job object.")
        handle_job = self._kernel32.CreateJobObject(None, job_name)
        if handle_job == 0:
            raise ctypes.WinError(ctypes.get_last_error())
        logger.debug(f"Job object: {handle_job:#x}")
        self._handle_job = handle_job

        # associate io port completion, if any
        if not self._handle_io_port:
            return
        job_completion_port = JOBOBJECT_ASSOCIATE_COMPLETION_PORT()
        job_completion_port.CompletionKey = self._handle_job
        job_completion_port.CompletionPort = self._handle_io_port
        self._set_information_job_object(
            job_completion_port,
            JobObjectAssociateCompletionPortInformation)

    def create_process(self, process_path: pathlib.Path,
                       command_line: Optional[str] = None) -> None:
        """Create a new process object to be associated with the main job.

        Args:
            process_path: The path to the main binary executable.
            command_line: The command line for the process.

        Raises:
            ctypes.WinError: An error occurred while trying to create the
                process.

        Notes:
            The command line is prepended with the full binary path.
        """
        # create the process with its main thread in a suspended state
        logger.debug("Creating suspended process.")
        si = STARTUPINFO()
        si.cb = ctypes.sizeof(STARTUPINFO)
        pi = PROCESS_INFORMATION()

        full_proc_path = process_path.resolve()
        cmd_line_str = str(full_proc_path) + (
            command_line if command_line is not None else "")

        cmd_line = self._kernel32.create_buffer(cmd_line_str)
        current_dir = self._kernel32.create_buffer(str(full_proc_path.parent))

        ret_val = self._kernel32.CreateProcess(
            None,  # lpApplicationName
            cmd_line,  # lpCommandLine
            None,  # lpProcessAttributes
            None,  # lpThreadAttributes
            False,  # bInheritHandles
            CREATE_SUSPENDED,  # dwCreationFlags
            None,  # lpEnvironment
            current_dir,  # lpCurrentDirectory
            ctypes.byref(si),  # lpStartupInfo
            ctypes.byref(pi),  # lpProcessInformation
        )
        if ret_val == 0:
            raise ctypes.WinError(ctypes.get_last_error())
        logger.debug(f"Process: {pi.hProcess:#x}; Thread: {pi.hThread:#x}")

        self._handle_process = pi.hProcess
        self._handle_thread = pi.hThread

    def get_process(self, pid: int) -> None:
        """Get a process object from its process identifier (PID). This process
        will be associated with the main job.

        Args:
            pid: The pid of the process to associate with the job.

        Raises:
            ctypes.WinError: An error occurred while trying to ge the process
                object form its PID. Ensure you have
                sufficient rights upon the process.
        """
        handle_process = self._kernel32.OpenProcess(
            PROCESS_SET_QUOTA | PROCESS_TERMINATE, False, pid)
        if not handle_process:
            raise ctypes.WinError(ctypes.get_last_error())
        logger.debug(f"Process: {handle_process:#x}")
        self._handle_process = handle_process

    def limit_process_memory(self, memory_limit: int) -> None:
        """Effectively limit the process memory that the target process can
            allocates.

        Args:
            memory_limit: The memory limit of the process, in MiB (MebiBytes).

        Raises:
            ctypes.WinError: There was an error while trying to limit the
                process memory.
            ValueError: The given memory limit is not valid or there's no job.
        """
        logger.info(f"Limiting process memory to {memory_limit} MiB "
                    f"({memory_limit * 1024 * 1024} bytes)")
        if memory_limit <= 0:
            raise ValueError(f"Memory limit can be 0 or negative; got: "
                             f"{memory_limit}")
        if not self._handle_job:
            raise ValueError("Job handle is NULL.")
        # query current job object
        job_info = JOBOBJECT_EXTENDED_LIMIT_INFORMATION()
        job_info = self._query_information_job_object(
            job_info, JobObjectExtendedLimitInformation)

        # limit the process memory; note: not the job memory!
        logger.debug("Setting job information.")
        job_info.BasicLimitInformation.LimitFlags |= \
            JOB_OBJECT_LIMIT_PROCESS_MEMORY
        # MiB to bytes
        job_info.ProcessMemoryLimit = memory_limit * 1024 * 1024
        self._set_information_job_object(
            job_info, JobObjectExtendedLimitInformation)

    def wait_for_job(self) -> bool:
        """Wait for the job completion. This function returns when the last
        process of the job exits.

        Notes:
            This function returns immediately if there's no IO ports.

        Returns:
            True if the function successfully waited for the job to finish,
            False otherwise.
        """
        logger.info("Waiting for the job.")
        if self.is_started_process:
            # resume the main thread, as we started the process ourselves.
            self._resume_main_thread()
        if not self.has_io_port:
            return False
        msg_map: Dict[int, str] = {
            JOB_OBJECT_MSG_END_OF_JOB_TIME: "End of job time",
            JOB_OBJECT_MSG_END_OF_PROCESS_TIME: "End of process time",
            JOB_OBJECT_MSG_ACTIVE_PROCESS_LIMIT:
                "Active process limit reached",
            JOB_OBJECT_MSG_ACTIVE_PROCESS_ZERO:
                "No more active process in job",
            JOB_OBJECT_MSG_NEW_PROCESS: "New process in job",
            JOB_OBJECT_MSG_EXIT_PROCESS: "A process in the job exited",
            JOB_OBJECT_MSG_ABNORMAL_EXIT_PROCESS:
                "A process in the job exited abnormally",
            JOB_OBJECT_MSG_PROCESS_MEMORY_LIMIT:
                "A process in the job reached its memory limit",
            JOB_OBJECT_MSG_JOB_MEMORY_LIMIT:
                "The job has reached its memory limit",
            JOB_OBJECT_MSG_NOTIFICATION_LIMIT:
                "The job reached a notification limit",
            JOB_OBJECT_MSG_JOB_CYCLE_TIME_LIMIT:
                "The CPU cycle limit for the job has been reached",
            JOB_OBJECT_MSG_SILO_TERMINATED: "A silo a terminated.",
        }

        return_val = False

        completion_code = DWORD(0)
        completion_key = ULONG_PTR(0)
        overlapped = OVERLAPPED()
        lp_overlapped = ctypes.POINTER(OVERLAPPED)(overlapped)
        while True:
            ret_val = self._kernel32.GetQueuedCompletionStatus(
                self._handle_io_port,
                ctypes.byref(completion_code),
                ctypes.byref(completion_key),
                ctypes.byref(lp_overlapped),
                INFINITE,
            )
            if ret_val == 0:
                # error from GetQueuedCompletionStatus
                logger.error(f"Error GetQueuedCompletionStatus: "
                             f"{ctypes.get_last_error():#x}")
                break
            if completion_key.value != self._handle_job:
                # we received an event, but it's not from our job; we can
                # ignore it!
                continue
            # message
            msg = msg_map.get(completion_code.value)
            if msg:
                logger.info(f"IO Port Message: {msg}")
            if completion_code.value == JOB_OBJECT_MSG_ACTIVE_PROCESS_ZERO:
                # no more processes in the job, we can exit.
                return_val = True
                break
        logger.info("Job wait finished")
        return return_val


def main(args: argparse.Namespace) -> int:
    # runs only on Windows
    if platform.system().lower() != "windows":
        logger.error("This script only works on Microsoft Windows systems.")
        return -1

    # check the memory limit
    if args.memory <= 0:
        logger.debug(f"Memory limit must be above 0, got: {args.memory}")
        return -1

    # check the command type
    if args.command_name == "process":
        # check the binary file exists
        if not args.process.is_file():
            logger.debug(f"The given file path '{args.process}' is not a file "
                         f"or doesn't exist.")
            return -1
        logger.info(f"Process Path: {args.process}; Memory limit (MiB): "
                    f"{args.memory}")
    elif args.command_name == "pid":
        # pid to int
        if args.pid.startswith("0x") or args.pid.startswith("0X"):
            base = 16
        else:
            base = 10
        args.pid = int(args.pid, base)
        logger.info(f"Process Pid: {args.pid:#x}; Memory limit (MiB): "
                    f"{args.memory}")
    else:
        logger.error(f"Unknown command: '{args.command}'")
        return -1

    with ProcessLimiter() as proc_limiter:
        proc_limiter.create_job("MEMORY_LIMITER_JOB")

        if args.command_name == "process":
            proc_limiter.create_process(args.process)
        else:
            proc_limiter.get_process(args.pid)
        proc_limiter.assign_process_to_job()
        proc_limiter.limit_process_memory(args.memory)
        if not proc_limiter.wait_for_job():
            input("Press <enter> to exit this script")

    print("Script done.")
    return 0


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(
        description="C:DDA Memory Limit test script.")

    #
    # options for all sub-commands
    #
    arg_parser.add_argument(
        "-l",
        "--log-level",
        choices=['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        default='INFO',
        help="Set the logging level.",
    )
    arg_parser.add_argument(
        "-m", "--memory", action="store", type=int, default=1024,
        help="Maximum process memory size in MiB."
    )

    #
    # sub-commands
    #
    subparsers = arg_parser.add_subparsers(help='help for sub-commands',
                                           dest="command_name")

    parser_new_process = subparsers.add_parser(
        'process', help='Start a new process.')
    parser_new_process.add_argument(
        "process", action="store", type=pathlib.Path,
        help="Path to process binary file.")

    parser_pid = subparsers.add_parser(
        'pid', help='Enforce limit on a given process pid.')
    parser_pid.add_argument(
        "pid", action="store", type=str,
        help="Process pid on which to enforce memory limit.")

    parsed_args = arg_parser.parse_args()

    if not parsed_args.command_name:
        arg_parser.error("'process' or 'pid' command is required.")
    logging_level = logging.getLevelName(parsed_args.log_level)
    logging.basicConfig(level=logging_level)
    logger.setLevel(logging_level)

    sys.exit(main(parsed_args))



#!/usr/bin/env python3
"""
  Usage:
    `cd path/to/Cataclysm-DDA/`
    `python3 tools/json_tools/update-translate-dialogue-mod.py`
    then lint via `json_formatter.cgi`, check `doc/JSON/JSON_STYLE.md`

  You can also add `"//": "mod_update_script_compact"`
  to dynamic lines to make the script concatenate that line
  more compactly and without newlines, which is useful if
  the original dialog also uses concatenation.
"""

import json


def transform_dynamic_line(dl, id):
    if type(dl) is str or \
       type(dl) is int or \
       type(dl) is bool or \
       type(dl) is None:
        return dl

    if type(dl) is list:
        out = []
        for line in dl:
            out.append(transform_dynamic_line(line, id))
        return out

    if type(dl) is dict:
        if ("//~" in dl) and ("str" in dl):
            if ("//" in dl) and (dl["//"] == "mod_update_script_compact"):
                return {"concatenate": [
                    dl["str"],
                    " [TRANSLATE: ",
                    dl["//~"],
                    "]"]}
            return {"concatenate": [
                dl["str"],
                "\"\n\n[TRANSLATE:] \"",
                dl["//~"]]}
        out = {}
        for k in dl:
            out[k] = transform_dynamic_line(dl[k], id)
        return out

    return dl


def parse_dialogue(obj, dialogues):
    if "dynamic_line" not in obj:
        return

    out = {}
    out["id"] = obj["id"]
    out["type"] = obj["type"]
    dl = obj["dynamic_line"]

    out["dynamic_line"] = transform_dynamic_line(dl, out["id"])

    if "TRANSLATE:" in json.dumps(out):
        dialogues.append(out)


def main():
    input_dir = "data/json/npcs/exodii/"
    output_dir = "data/mods/translate-dialogue/"
    files = ["exodii_merchant_talk.json",
             "exodii_merchant_talk_exodization.json"]

    for f in files:
        dialogues = []
        with open(input_dir + f, encoding="utf-8") as fp:
            print(f"Reading from {input_dir + f}")
            json_data = json.load(fp)

        for obj in json_data:
            try:
                parse_dialogue(obj, dialogues)
            except Exception as err:
                print(f"Error: {err}")

        if dialogues:
            with open(output_dir + f, "w", encoding="utf-8") as fp:
                print(f"  Writing to {output_dir + f}")
                json.dump(dialogues, fp, ensure_ascii=False, indent="  ")


main()



from ..write_text import write_text


def parse_scenario(json, origin):
    name = ""
    if "name" in json:
        name = json["name"]
        write_text(name, origin, context="scenario_male",
                   comment="Scenario name for male")
        write_text(name, origin, context="scenario_female",
                   comment="Scenario name for female")
    if name == "":
        name = json["id"]

    if "description" in json:
        write_text(json["description"], origin, context="scen_desc_male",
                   comment="Description of scenario \"{}\" for male"
                   .format(name))
        write_text(json["description"], origin, context="scen_desc_female",
                   comment="Description of scenario \"{}\" for female"
                   .format(name))

    if "start_name" in json:
        write_text(json["start_name"], origin, context="start_name",
                   comment="Starting location of scenario \"{}\""
                   .format(name))



from ..helper import get_singular_name
from ..write_text import write_text
from .monster_attack import parse_monster_attack


def parse_monster_concrete(json, origin, name):
    if "description" in json:
        write_text(json["description"], origin, c_format=False,
                   comment="Description of monster \"{}\"".format(name))

    if "death_function" in json:
        if "message" in json["death_function"]:
            write_text(json["death_function"]["message"], origin,
                       comment="Death message of monster \"{}\"".format(name))

    if "special_attacks" in json:
        for attack in json["special_attacks"]:
            if "type" in attack and attack["type"] == "monster_attack":
                parse_monster_attack(attack, origin)
                continue
            if "description" in attack:
                write_text(attack["description"], origin,
                           comment="Description of special attack of monster "
                           "\"{}\"".format(name))
            if "monster_message" in attack:
                write_text(attack["monster_message"], origin,
                           comment="Monster message of special attack of "
                           "monster \"{}\"".format(name))
            if "targeting_sound" in attack:
                write_text(attack["targeting_sound"], origin,
                           comment="Targeting sound of special attack of "
                           "monster \"{}\"".format(name))
            if "no_ammo_sound" in attack:
                write_text(attack["no_ammo_sound"], origin,
                           comment="No ammo sound of special attack of "
                           "monster \"{}\"".format(name))

    if "weakpoints" in json:
        for weakpoint in json["weakpoints"]:
            if "name" in weakpoint:
                write_text(weakpoint["name"], origin,
                           comment="Sentence fragment describing a "
                           "weakpoint for monster \"{}\"".format(name))

    if "petfood" in json:
        if "feed" in json["petfood"]:
            write_text(json["petfood"]["feed"], origin,
                       comment="Feed message of monster \"{}\"".format(name))
        if "pet" in json["petfood"]:
            write_text(json["petfood"]["pet"], origin,
                       comment="Pet message of monster \"{}\"".format(name))


def parse_monster(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Monster name", plural=True)
    elif "id" in json:
        name = json["id"]

    parse_monster_concrete(json, origin, name)

    if "extend" in json:
        parse_monster_concrete(json["extend"], origin, name)



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify json directory")
args_dict = vars(args.parse_args())

inheriting_item_ids = []


def gen_new(path):
    change = False
    with open(path, "r", encoding="utf-8") as json_file:
        json_data = json.load(json_file)
        for jo in json_data:
            if "capacity" in jo:
                if "ammo_type" in jo:
                    pocket_data = [{
                        "pocket_type": "MAGAZINE",
                        "ammo_restriction": {k: jo["capacity"]
                                             for k in jo["ammo_type"]}
                    }]
                    jo["pocket_data"] = pocket_data
                    change = True
                else:
                    inheriting_item_ids.append(jo["id"])

    if change:
        return json_data
    else:
        return None


for root, directories, filenames in os.walk(args_dict["dir"]):
    for filename in filenames:
        path = os.path.join(root, filename)
        if path.endswith(".json"):
            new = gen_new(path)
            if new:
                with open(path, "w", encoding="utf-8") as jf:
                    json.dump(new, jf, ensure_ascii=False)
                os.system(f"./tools/format/json_formatter.cgi {path}")

print(*inheriting_item_ids, sep="\n")



#!/usr/bin/env python3

import os
import sys
import util

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))
GG_DIR = os.path.normpath(os.path.join(
    SCRIPT_DIR, '..', '..', 'data', 'mods', 'Generic_Guns'))

# We might want to also warn about grenades and rockets, but leaving them
# whitelisted for now.
AMMO_TYPE_WHITELIST = {
    '40x46mm',  # Grenade
    'atgm',  # Rocket
    'atlatl',
    'bolt_ballista',
    'barb',
    'battery',
    'BB',
    'blunderbuss',
    'bolt',
    'cannon',
    'charcoal',
    'chemical_spray',
    'fishspear',
    'flammable',
    'gene_sting',
    'nether_huntsman_javelin_ammo',
    'm235',  # Rocket
    'metal_rail',
    'nail',
    'nuts_bolts',
    'oxygen',
    'paintball',
    'pebble',
    'plasma',
    'rock',
    'signal_flare',
    'methanol_fuelcell',
    'weldgas'
}
SKILL_WHITELIST = {
    'archery',
    'launcher',
}
# Whitelisted specific ids are mostly obsolete stuff.
# This could go away if obsolete stuff were explicitly marked as such.
ID_WHITELIST = {
    # Guns
    'coilgun',
    'slamfire_shotgun',
    'slamfire_shotgun_d',
    'ftk93',
    'l_bak_223',
    'pneumatic_shotgun',
    'rifle_223',
    'ksg-25',
    'raging_judge',
    'american_180',
    'gene_sting_gun',
    'nether_huntsman_arm',
    'ppsh',
    'af2011a1_38super',
    'm26_mass',
    # Magazines
    'a180mag',
    'a180mag1',
    'a180mag2',
    'a180mag3',
    'a180mag4',
    'af2011a1mag',
    '454_speedloader6',
    '223_speedloader5',
    'coin_wrapper',
    'exodiisapramag5',
    'robofac_gun_40mm_3rd',
    'robofac_gun_40mm_5rd',
    'robofac_gun_40mm_10rd',
    'bio_shotgun_gun',
    'gasfilter_med',
    'gasfilter_sm',
    'matchhead_30carbine',
    'matchhead_30carbine_jsp',
    "rebreather_cartridge",
    "rebreather_cartridge_air",
    "rebreather_cartridge_o2",
    "rebreather_cartridge_regen"
}


def items_of_type(data, type):
    result = []
    for i in data:
        if 'type' not in i:
            dump = util.CDDAJSONWriter(i).dumps()
            print("json entry has no 'type' field: " + dump)
            sys.exit(1)
        if i['type'] == type:
            result.append(i)
    return result


def items_of_subtype(data, subtype):
    result = []
    for i in data:
        if 'type' not in i:
            dump = util.CDDAJSONWriter(i).dumps()
            print("json entry has no 'type' field: " + dump)
            sys.exit(1)
        if i['type'] == "ITEM":
            if 'subtypes' in i:
                if subtype in i['subtypes']:
                    result.append(i)
    return result


def get_ids(items):
    result = set()
    for i in items:
        id = i.get('id', [])
        if isinstance(id, str):
            result.add(id)
        else:
            result.update(id)
    return result


def get_id(i):
    if 'id' in i:
        return i['id']
    if 'abstract' in i:
        return i['abstract']
    raise RuntimeError('item lacks id: %r' % i)


def get_ancestors(items_by_id, item):
    result = [item]
    while 'copy-from' in item and item['copy-from'] in items_by_id:
        item = items_by_id[item['copy-from']]
        if item in result:
            raise RuntimeError(
                'Cyclic dependency in copy-from involving %s' %
                item['copy-from'])
        result.append(item)
    return result


def items_for_which_any_ancestor(items, pred):
    items_by_id = {get_id(i): i for i in items}
    result = []
    for i in items:
        if any(pred(ancestor) for ancestor in get_ancestors(items_by_id, i)):
            result.append(i)
    return result


def items_for_which_all_ancestors(items, pred):
    items_by_id = {get_id(i): i for i in items}
    result = []
    for i in items:
        if all(pred(ancestor) for ancestor in get_ancestors(items_by_id, i)):
            result.append(i)
    return result


def blacklisted_items(data):
    blacklists = items_of_type(data, 'ITEM_BLACKLIST')
    items = set()
    for blacklist in blacklists:
        if 'items' in blacklist:
            items.update(set(blacklist['items']))
    return items


def main():
    core_data, core_errors = util.import_data()
    print('Importing Generic Guns data from %r' % GG_DIR)
    gg_data, gg_errors = util.import_data(GG_DIR)

    if core_errors or gg_errors:
        print('Errors reading json:\n%s' % '\n'.join(core_errors + gg_errors))
        sys.exit(1)

    gg_migrations = get_ids(items_of_type(gg_data, 'MIGRATION'))
    gg_blacklist = blacklisted_items(gg_data)

    core_guns = items_of_subtype(core_data, 'GUN')

    def is_not_fake_item(i):
        return i.get('copy-from', '') != 'fake_item'

    def is_not_whitelisted_skill(i):
        return 'skill' in i and i['skill'] not in SKILL_WHITELIST

    def has_pockets(i):
        return 'pocket_data' in i

    def lacks_whitelisted_pocket(i):
        return not any(
            pocket.get('ammo_restriction', {}).keys() & AMMO_TYPE_WHITELIST
            for pocket in i.get('pocket_data', []))

    def can_be_unwielded(i):
        return 'NO_UNWIELD' not in i.get('flags', [])

    core_guns = items_for_which_all_ancestors(
        core_guns, is_not_fake_item)
    core_guns = items_for_which_any_ancestor(
        core_guns, is_not_whitelisted_skill)
    core_guns = items_for_which_any_ancestor(
        core_guns, has_pockets)
    core_guns = items_for_which_all_ancestors(
        core_guns, lacks_whitelisted_pocket)
    core_guns = items_for_which_all_ancestors(
        core_guns, can_be_unwielded)

    core_magazines = items_of_subtype(core_data, 'MAGAZINE')
    core_magazines = items_for_which_all_ancestors(
        core_magazines, lacks_whitelisted_pocket)

    core_ammo = items_of_subtype(core_data, 'AMMO')

    def is_not_whitelisted_ammo_type(i):
        return 'ammo_type' in i and i['ammo_type'] not in AMMO_TYPE_WHITELIST

    def is_bullet(i):
        return i.get('damage', {}).get('damage_type', '') == 'bullet'

    core_bullets = items_for_which_any_ancestor(core_ammo, is_bullet)
    core_bullets = items_for_which_any_ancestor(
        core_bullets, is_not_whitelisted_ammo_type)

    if (not gg_migrations or not core_guns or not core_magazines or
            not core_ammo):
        print('One of the collections is empty; something has gone wrong with '
              'data collection')
        return 1

    returncode = 0

    def check_missing(items, name):
        ids = get_ids(items) - ID_WHITELIST - gg_blacklist

        missing_migrations = ids - gg_migrations
        if missing_migrations:
            print('Missing Generic Guns migrations for these types of %s:' %
                  name)
            print('\n'.join(sorted(missing_migrations)))
            print()
            nonlocal returncode
            returncode = 1

    check_missing(core_bullets, 'ammo')
    check_missing(core_magazines, 'magazine')
    check_missing(core_guns, 'guns')

    if returncode:
        print('The above errors can be resolved by either adding suitable '
              'migrations to Generic Guns or adding to the whitelists of '
              'things not requiring migration in '
              'tools/json_tools/generic_guns_validator.py')

    return returncode


if __name__ == '__main__':
    sys.exit(main())



from ..write_text import write_text


def parse_recipe_category(json, origin):
    cid = json["id"]
    if cid == 'CC_NONCRAFT':
        return
    cat_name = cid.split("_")[1]
    write_text(cat_name, origin, comment="Crafting recipes category name")

    for subcat in json.get("recipe_subcategories", []):
        if subcat == 'CSC_ALL':
            write_text("ALL", origin,
                       comment="Crafting recipes subcategory \"all\"")
        else:
            subcat_name = subcat.split('_')[2]
            write_text(subcat_name, origin,
                       comment="Crafting recipes subcategory of \"{}\" cat."
                       .format(cat_name))



#!/usr/bin/env python3

# The goal of this script is to print out sets of mods for testing.  Each line
# of output is a comma-separated list of mods.  Together the lines should cover
# all in-repo mods, in as few lines as possible.  Each line must contain only
# mods which are mutually compatible.

import glob
import json
import os

mods_this_time = []

exclusions = [
    # Tuple of (mod_id, mod_id) - these two mods will be incompatible
    # Note that mod_id is case sensitive
]


def compatible_with(mod, existing_mods):
    if mod in total_conversions and total_conversions & set(existing_mods):
        return False
    for entry in existing_mods:
        if mod in all_mod_conflicts[entry] or entry in all_mod_conflicts[mod]:
            return False
    for entry in exclusions:
        if entry[0] == mod and entry[1] in existing_mods:
            return False
        if entry[1] == mod and entry[0] in existing_mods:
            return False
    return True


def add_mods(mods):
    for mod in mods:
        if mod not in all_mod_dependencies:
            # Either an invalid mod id, or blacklisted.
            return False
    for mod in mods:
        if mod in mods_this_time:
            continue
        if not compatible_with(mod, mods_this_time):
            return False
        if add_mods(all_mod_dependencies[mod]):
            mods_this_time.append(mod)
        else:
            return False
    return True


def print_modlist(modlist, master_list):
    print(','.join(modlist))
    master_list -= set(modlist)
    modlist.clear()


all_mod_dependencies = {}
all_mod_conflicts = {}
total_conversions = set()
obsolete_mods = set()

for info in glob.glob('data/mods/*/modinfo.json'):
    mod_info = json.load(open(info, encoding='utf-8'))
    for e in mod_info:
        if (e["type"] == "MOD_INFO"):
            ident = e["id"]
            if ("obsolete" in e and e["obsolete"]):
                obsolete_mods.add(ident)
                continue
            all_mod_dependencies[ident] = e.get("dependencies", [])
            all_mod_conflicts[ident] = e.get("conflicts", [])
            if e["category"] == "total_conversion":
                total_conversions.add(ident)

for r, d, f in os.walk('data/mods'):
    if 'mod_interactions' not in d:
        continue
    if 'modinfo.json' not in f:
        continue
    ident = ""
    info_path = os.path.join(r, 'modinfo.json')
    mod_info = json.load(open(info_path, encoding='utf-8'))
    for e in mod_info:
        if (e["type"] == "MOD_INFO" and
                ("obsolete" not in e or not e["obsolete"])):
            ident = e["id"]
    if ident == "":
        continue
    add_mods([ident])
    for mod in os.scandir(os.path.join(r, 'mod_interactions')):
        ident = os.path.basename(mod.path)
        if ident in obsolete_mods:
            continue
        mods_this_time.append(os.path.basename(mod.path))
    print(','.join(mods_this_time))
    mods_this_time.clear()

mods_remaining = set(all_mod_dependencies)

print_modlist(mods_this_time, mods_remaining)

while mods_remaining:
    for mod in mods_remaining:
        if mod not in mods_this_time:
            add_mods([mod])
    if not mods_remaining & set(mods_this_time):
        raise RuntimeError(
            'mods remain ({}) but none could be added'.format(mods_remaining))
    print_modlist(mods_this_time, mods_remaining)



from .effect import parse_effect_on_condition
from ..write_text import write_text


use_action_msg_keys = [
    "activate_msg",
    "activation_message",
    "auto_extinguish_message",
    "bury_question",
    "charges_extinguish_message",
    "deactive_msg",
    "descriptions",
    "description",
    "done_message",
    "failure_message",
    "friendly_msg",
    "gerund",
    "holster_msg",
    "holster_prompt",
    "hostile_msg",
    "lacks_fuel_message",
    "menu_text",
    "message",
    "msg",
    "need_charges_msg",
    "need_fire_msg",
    "no_deactivate_msg",
    "noise_message",
    "non_interactive_msg",
    "not_ready_msg",
    "out_of_power_msg",
    "sound_message",
    "success_message",
    "unfold_msg",
    "use_message",
    "verb",
    "voluntary_extinguish_message",
    "water_extinguish_message",
]


def parse_use_action(json, origin, item_name):
    if type(json) is dict:
        for msg_key in use_action_msg_keys:
            if msg_key in json:
                write_text(json[msg_key], origin,
                           comment="\"{0}\" action message of item \"{1}\""
                           .format(msg_key, item_name))
        if json["type"] == "place_trap" and "bury" in json:
            write_text(json["bury"]["done_message"], origin,
                       comment="bury trap message of item \"{}\""
                       .format(item_name))
        if (json["type"] == "effect_on_conditions" and
                "effect_on_conditions" in json):
            for e in json["effect_on_conditions"]:
                if type(e) is dict:
                    parse_effect_on_condition(e, origin,
                                              "use action of item \"{}\""
                                              .format(item_name))
    elif type(json) is list:
        for use_action in json:
            parse_use_action(use_action, origin, item_name)



from ..write_text import write_text


def parse_camp_name(overmap, origin):
    if "camp_name" in overmap:
        write_text(overmap["camp_name"], origin,
                   comment="Name of NPC faction camp")


def parse_overmap_special(json, origin):
    if "subtype" in json and json["subtype"] == "mutable":
        for mutable_overmap, overmap in json["overmaps"].items():
            parse_camp_name(overmap, origin)
    else:
        for overmap in json["overmaps"]:
            parse_camp_name(overmap, origin)



#!/usr/bin/env python3
'''
Merge all tile entries and PNGs in a compositing tileset directory into
a tile_config.json and tilesheet .png file(s) ready for use in CDDA.

Examples:

    %(prog)s ../CDDA-Tilesets/gfx/Retrodays/
    %(prog)s --use-all ../CDDA-Tilesets/gfx/UltimateCataclysm/

By default, output is written back to the source directory. Pass an output
directory as the last argument to place output files there instead. The
output directory will be created if it does not already exist.
'''

import argparse
import json
import logging
import os
import subprocess
import sys

from logging.config import dictConfig
from pathlib import Path
from typing import Any, Optional, Tuple, Union

try:
    vips_path = os.getenv("LIBVIPS_PATH")
    if vips_path is not None and vips_path != "":
        os.environ["PATH"] += ";" + os.path.join(vips_path, "bin")
    import pyvips
    Vips = pyvips
except ImportError:
    import gi
    gi.require_version('Vips', '8.0')  # NoQA
    from gi.repository import Vips


# variable for silent script run (no output to terminal)
run_silent = True

# variable for progress bar support (tqdm module dependency)
no_tqdm = False

# File name to ignore containing directory
ignore_file = ".scratch"


# progress bar setup
# requires tqdm module, if not installed prompts
# to ask permission to install it via PIP
# it also can work without it, in 'CONCISE mode',
# where output is limited to stages of processing
def setup_progress_bar() -> str:
    global no_tqdm
    global tqdm
    try:
        from tqdm import tqdm
        no_tqdm = False
        return 'VERBOSE'
    except ImportError:
        from tkinter.messagebox import askyesno
        txt_title = "compose.py: TQDM module not installed"
        txt_message = "VERBOSE mode requires TQDM module"
        " to display progress bar(s). "
        "Do you want to install TQDM "
        "(and it's dependencies) via PIP?"
        if askyesno(txt_title, txt_message):
            try:
                sub = subprocess.Popen([sys.executable,
                                       '-m', 'pip', 'install', "tqdm"],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT)
                print(sub.communicate()[0].decode("utf-8"))
                from tqdm import tqdm
                no_tqdm = False
                return 'VERBOSE'
            except ImportError:  # still no go, fall back to CONCISE mode
                print("Could not install/import TQDM module."
                      " Display in CONCISE mode instead.")
                no_tqdm = True
                return 'CONCISE'
        else:  # fall back to CONCISE mode
            print("Display is in CONCISE mode.")
            no_tqdm = True
            return 'CONCISE'


class LevelTrackingFilter(logging.Filter):
    """
    Logging handler that will remember the highest level that was called
    """
    def __init__(self):
        super().__init__()
        self.level = logging.NOTSET

    def filter(self, record):
        self.level = max(self.level, record.levelno)
        return True


LOGGING_CONFIG = {
    'formatters': {
        'standard': {'format': '%(levelname)s %(funcName)s: %(message)s'},
    },
    'handlers': {
        'default': {
            'level': 'NOTSET',  # will be set later
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        __name__: {
            'handlers': ['default'],
            'level': 'INFO',
        },
    },
    'disable_existing_loggers': False,
    'version': 1,
}

PROPERTIES_FILENAME = 'tileset.txt'

PNGSAVE_ARGS = {
    'compression': 9,
    'strip': True,
    'filter': 8,
}

FALLBACK = {
    "file": "fallback.png",
    "tiles": [],
    "ascii": [
        {"offset": 0, "bold": False, "color": "BLACK"},
        {"offset": 256, "bold": True, "color": "WHITE"},
        {"offset": 512, "bold": False, "color": "WHITE"},
        {"offset": 768, "bold": True, "color": "BLACK"},
        {"offset": 1024, "bold": False, "color": "RED"},
        {"offset": 1280, "bold": False, "color": "GREEN"},
        {"offset": 1536, "bold": False, "color": "BLUE"},
        {"offset": 1792, "bold": False, "color": "CYAN"},
        {"offset": 2048, "bold": False, "color": "MAGENTA"},
        {"offset": 2304, "bold": False, "color": "YELLOW"},
        {"offset": 2560, "bold": True, "color": "RED"},
        {"offset": 2816, "bold": True, "color": "GREEN"},
        {"offset": 3072, "bold": True, "color": "BLUE"},
        {"offset": 3328, "bold": True, "color": "CYAN"},
        {"offset": 3584, "bold": True, "color": "MAGENTA"},
        {"offset": 3840, "bold": True, "color": "YELLOW"}
    ]
}


log = logging.getLogger(__name__)


class FailFastHandler(logging.StreamHandler):
    def emit(self, record):
        sys.exit(1)


def write_to_json(
    pathname: str,
    data: Union[dict, list],
    format_json: bool = False,
) -> None:
    '''
    Write data to a JSON file
    '''
    kwargs = {
        'ensure_ascii': False,
    }
    if format_json:
        kwargs['indent'] = 2

    with open(pathname, 'w', encoding="utf-8") as file:
        json.dump(data, file, **kwargs)

    if not format_json:
        return

    json_formatter = Path('tools/format/json_formatter.cgi')
    if json_formatter.is_file():
        cmd = [json_formatter, pathname]
        subprocess.call(cmd)
    else:
        log.warning(
            '%s not found, Python built-in formatter was used.',
            json_formatter)


def list_or_first(iterable: list) -> Any:
    '''
    Strip unneeded container list if there is only one value
    '''
    return iterable[0] if len(iterable) == 1 else iterable


def read_properties(filepath: str) -> dict:
    '''
    tileset.txt reader
    '''
    with open(filepath, 'r', encoding="utf-8") as file:
        pairs = {}
        for line in file.readlines():
            line = line.strip()
            if line and not line.startswith('#'):
                key, value = line.split(':')
                pairs[key.strip()] = value.strip()
    return pairs


class ComposingException(Exception):
    '''
    Base class for all composing exceptions
    '''


class Tileset:
    '''
    Referenced sprites memory and handling, tile entries conversion
    '''
    def __init__(
        self,
        source_dir: Path,
        output_dir: Path,
        use_all: bool = False,
        obsolete_fillers: bool = False,
        palette_copies: bool = False,
        palette: bool = False,
        format_json: bool = False,
        only_json: bool = False,
    ) -> None:
        self.source_dir = source_dir
        self.output_dir = output_dir
        self.use_all = use_all
        self.obsolete_fillers = obsolete_fillers
        self.palette_copies = palette_copies
        self.palette = palette
        self.format_json = format_json
        self.only_json = only_json
        self.output_conf_file = None

        self.pngnum = 0
        self.unreferenced_pngnames = {
            'main': [],
            'filler': [],
        }

        self.pngname_to_pngnum = {'null_image': 0}

        if not self.source_dir.is_dir() or \
                not os.access(self.source_dir, os.R_OK):
            raise ComposingException(
                f'Error: cannot open directory {self.source_dir}')

        self.processed_ids = []
        info_path = self.source_dir.joinpath('tile_info.json')
        self.sprite_width = 16
        self.sprite_height = 16
        self.zlevel_height = 0
        self.pixelscale = 1
        self.iso = False
        self.retract_dist_min = -1.0
        self.retract_dist_max = 1.0
        self.info = [{}]

        if not os.access(info_path, os.R_OK):
            raise ComposingException(f'Error: cannot open {info_path}')

        with open(info_path, 'r', encoding="utf-8") as file:
            self.info = json.load(file)
            self.sprite_width = self.info[0].get('width', self.sprite_width)
            self.sprite_height = self.info[0].get('height', self.sprite_height)
            self.zlevel_height = self.info[0].get('zlevel_height',
                                                  self.zlevel_height)
            self.pixelscale = self.info[0].get('pixelscale', self.pixelscale)
            self.retract_dist_min = self.info[0].get('retract_dist_min',
                                                     self.retract_dist_min)
            self.retract_dist_max = self.info[0].get('retract_dist_max',
                                                     self.retract_dist_max)
            self.iso = self.info[0].get('iso', self.iso)

    def determine_conffile(self) -> str:
        '''
        Read JSON value from tileset.txt
        '''
        properties = {}

        for candidate_path in (self.source_dir, self.output_dir):
            properties_path = candidate_path.joinpath(PROPERTIES_FILENAME)
            if os.access(properties_path, os.R_OK):
                properties = read_properties(properties_path)
                if properties:
                    break

        if not properties:
            raise ComposingException(f'No valid {PROPERTIES_FILENAME} found')

        conf_filename = properties.get('JSON', None)

        if not conf_filename:
            raise ComposingException(
                f'No JSON key found in {PROPERTIES_FILENAME}')

        self.output_conf_file = conf_filename
        return self.output_conf_file

    def compose(self) -> None:
        '''
        Convert a composing tileset into a package readable by the game
        '''
        self.output_dir.mkdir(parents=True, exist_ok=True)
        tileset_confpath = self.output_dir.joinpath(self.determine_conffile())
        typed_sheets = {
            'main': [],
            'filler': [],
            'fallback': [],
        }
        fallback_name = 'fallback.png'

        # loop through tilesheets and parse all configs in subdirectories,
        # create sheet images
        added_first_null = False
        for config in self.info[1:]:
            sheet = Tilesheet(self, config)

            if not added_first_null:
                sheet.sprites.append(sheet.null_image)
                added_first_null = True

            if sheet.is_filler:
                sheet_type = 'filler'
            elif sheet.is_fallback:
                sheet_type = 'fallback'
            else:
                sheet_type = 'main'

            log.info('parsing %s tilesheet %s', sheet_type, sheet.name)
            if not run_silent:
                print("Composing [" + sheet_type +
                      "] tilesheet [" + sheet.name + "]...",
                      end=' ' if no_tqdm else '\n', flush=True)
            if sheet_type != 'fallback':
                sheet.walk_dirs()
                # TODO: generate JSON first
                # then create sheets if there are no errors
                if no_tqdm and not run_silent:
                    print("done.", flush=True)
                if not run_silent:
                    print("Saving output PNG...", end=' ', flush=True)
                # write output PNGs
                if not sheet.write_composite_png():
                    continue
                if not run_silent:
                    print("done.", flush=True)
                sheet.max_index = self.pngnum

            typed_sheets[sheet_type].append(sheet)

        # combine config data in the correct order
        sheet_configs = typed_sheets['main'] + typed_sheets['filler'] \
            + typed_sheets['fallback']

        # prepare "tiles-new", but remember max index of each sheet in keys
        tiles_new_dict = dict()

        def create_tile_entries_for_unused(
            unused: list,
            fillers: bool,
        ) -> None:
            # the list must be empty without use_all
            mode = unused if no_tqdm or run_silent else tqdm(unused)
            for unused_png in mode:
                if unused_png in self.processed_ids:
                    if not fillers:
                        log.warning(
                            '%(1)s sprite was not mentioned in any tile '
                            'entry but there is a tile entry for the %(1)s ID',
                            {'1': unused_png})
                    if fillers and self.obsolete_fillers:
                        log.warning(
                            'there is a tile entry for %s '
                            'in a non-filler sheet',
                            unused_png)
                    continue
                unused_num = self.pngname_to_pngnum[unused_png]
                sheet_min_index = 0
                for sheet_max_index, sheet_data in tiles_new_dict.items():
                    if sheet_min_index < unused_num <= sheet_max_index:
                        sheet_data['tiles'].append({
                            'id': unused_png,
                            'fg': unused_num,
                        })
                        self.processed_ids.append(unused_png)
                        break
                    sheet_min_index = sheet_max_index

        main_finished = False

        for sheet in sheet_configs:
            if sheet.is_fallback:
                fallback_name = sheet.name
                if not sheet.is_standard():
                    FALLBACK['sprite_width'] = sheet.sprite_width
                    FALLBACK['sprite_height'] = sheet.sprite_height
                    FALLBACK['sprite_offset_x'] = sheet.offset_x
                    FALLBACK['sprite_offset_y'] = sheet.offset_y
                    if sheet.offset_x_retracted != sheet.offset_x \
                            or sheet.offset_y_retracted != sheet.offset_y:
                        FALLBACK['sprite_offset_x_retracted'] = \
                            sheet.offset_x_retracted
                        FALLBACK['sprite_offset_y_retracted'] = \
                            sheet.offset_y_retracted
                    if sheet.pixelscale != 1.0:
                        FALLBACK['pixelscale'] = sheet.pixelscale
                continue
            if sheet.is_filler and not main_finished:
                create_tile_entries_for_unused(
                    self.handle_unreferenced_sprites('main'),
                    fillers=False
                )
                main_finished = True
            sheet_entries = []

            for tile_entry in sheet.tile_entries:
                # TODO: pop?
                converted_tile_entry = tile_entry.convert()
                if converted_tile_entry:
                    sheet_entries.append(converted_tile_entry)

            sheet_conf = {
                'file': sheet.name,
                '//': f'range {sheet.first_index} to {sheet.max_index}',
            }

            if not sheet.is_standard():
                sheet_conf['sprite_width'] = sheet.sprite_width
                sheet_conf['sprite_height'] = sheet.sprite_height
                sheet_conf['sprite_offset_x'] = sheet.offset_x
                sheet_conf['sprite_offset_y'] = sheet.offset_y
                if sheet.offset_x_retracted != sheet.offset_x \
                        or sheet.offset_y_retracted != sheet.offset_y:
                    sheet_conf['sprite_offset_x_retracted'] = \
                        sheet.offset_x_retracted
                    sheet_conf['sprite_offset_y_retracted'] = \
                        sheet.offset_y_retracted
                if sheet.pixelscale != 1.0:
                    sheet_conf['pixelscale'] = sheet.pixelscale

            sheet_conf['tiles'] = sheet_entries

            tiles_new_dict[sheet.max_index] = sheet_conf

        if not main_finished:
            create_tile_entries_for_unused(
                self.handle_unreferenced_sprites('main'),
                fillers=False,
            )

        create_tile_entries_for_unused(
            self.handle_unreferenced_sprites('filler'),
            fillers=True,
        )

        # finalize "tiles-new" config
        tiles_new = list(tiles_new_dict.values())

        FALLBACK['file'] = fallback_name
        tiles_new.append(FALLBACK)
        output_conf = {
            'tile_info': [{
                'pixelscale': self.pixelscale,
                'width': self.sprite_width,
                'height': self.sprite_height,
                'zlevel_height': self.zlevel_height,
                'iso': self.iso,
                'retract_dist_min': self.retract_dist_min,
                'retract_dist_max': self.retract_dist_max
            }],
            'tiles-new': tiles_new
        }

        # save the config
        write_to_json(tileset_confpath, output_conf, self.format_json)
        if no_tqdm and not run_silent:
            print("done.", flush=True)

    def handle_unreferenced_sprites(
        self,
        sheet_type: str,
    ) -> list:
        '''
        Either warn about unused sprites or return the list
        '''
        if self.use_all:
            return self.unreferenced_pngnames[sheet_type]

        for pngname in self.unreferenced_pngnames[sheet_type]:
            if pngname in self.processed_ids:
                log.error(
                    '%(1)s.png not used when %(1)s ID '
                    'is mentioned in a tile entry',
                    {'1': pngname})

            else:
                log.warning(
                    'sprite filename %s was not used in any %s %s entries',
                    pngname, sheet_type, self.output_conf_file)
        return []


class Tilesheet:
    '''
    Tilesheet reading and compositing
    '''
    def __init__(
        self,
        tileset: Tileset,
        config: dict,
    ) -> None:
        self.tileset = tileset

        self.name = next(iter(config))
        specs = config.get(self.name, {})

        self.sprite_width = specs.get(
            'sprite_width', tileset.sprite_width)
        self.sprite_height = specs.get(
            'sprite_height', tileset.sprite_height)
        self.offset_x = specs.get('sprite_offset_x', 0)
        self.offset_y = specs.get('sprite_offset_y', 0)
        self.offset_x_retracted = \
            specs.get('sprite_offset_x_retracted', self.offset_x)
        self.offset_y_retracted = \
            specs.get('sprite_offset_y_retracted', self.offset_y)

        self.pixelscale = specs.get('pixelscale', 1.0)

        self.sprites_across = specs.get('sprites_across', 16)
        self.exclude = specs.get('exclude', tuple())

        self.is_fallback = specs.get('fallback', False)
        self.is_filler = not self.is_fallback \
            and specs.get('filler', False)

        output_root = self.name.split('.png')[0]
        dir_name = \
            f'pngs_{output_root}_{self.sprite_width}x{self.sprite_height}'
        self.subdir_path = tileset.source_dir.joinpath(dir_name)

        self.output = tileset.output_dir.joinpath(self.name)

        self.tile_entries = []
        self.null_image = \
            Vips.Image.grey(self.sprite_width, self.sprite_height)
        self.sprites = []

        self.first_index = self.tileset.pngnum + 1
        self.max_index = self.tileset.pngnum

    def is_standard(self) -> bool:
        '''
        Check whether output object needs a non-standard size or offset config
        '''
        if self.offset_x or self.offset_y:
            return False
        if self.offset_x_retracted != self.offset_x \
                or self.offset_y_retracted != self.offset_y:
            return False
        if self.sprite_width != self.tileset.sprite_width:
            return False
        if self.sprite_height != self.tileset.sprite_height:
            return False
        if self.pixelscale != 1.0:
            return False
        return True

    def walk_dirs(self) -> None:
        '''
        Find and process all JSON and PNG files within sheet directory
        '''

        def filtered_tree(excluded):
            for root, dirs, filenames in \
                    os.walk(self.subdir_path, followlinks=True):
                # replace dirs in-place to prevent walking down excluded paths
                dirs[:] = [
                    d
                    for d in dirs
                    if Path(root).joinpath(d) not in excluded and
                    not Path(root).joinpath(d, ignore_file).is_file()
                ]
                yield [root, dirs, filenames]

        sorted_files = sorted(
            filtered_tree(list(map(self.subdir_path.joinpath, self.exclude))),
            key=lambda d: d[0]
        )
        mode = sorted_files if no_tqdm or run_silent else tqdm(sorted_files)
        for subdir_fpath, dirs, filenames in mode:
            subdir_fpath = Path(subdir_fpath)
            for filename in sorted(filenames):
                filepath = subdir_fpath.joinpath(filename)

                if filepath.suffixes == ['.png']:
                    self.process_png(filepath)

                elif filepath.suffixes == ['.json']:
                    self.process_json(filepath)

    def process_png(
        self,
        filepath: Path,
    ) -> None:
        '''
        Verify image root name is unique, load it and register
        '''
        if filepath.stem in self.tileset.pngname_to_pngnum:
            if not self.is_filler:
                log.error(
                    'duplicate root name %s: %s',
                    filepath.stem, filepath)

            if self.is_filler and self.tileset.obsolete_fillers:
                log.warning(
                    'root name %s is already present in a non-filler sheet: '
                    '%s',
                    filepath.stem, filepath)

            return

        if not self.tileset.only_json:
            self.sprites.append(self.load_image(filepath))
        else:
            self.sprites.append(None)

        self.tileset.pngnum += 1
        self.tileset.pngname_to_pngnum[filepath.stem] = self.tileset.pngnum
        self.tileset.unreferenced_pngnames[
            'filler' if self.is_filler else 'main'].append(filepath.stem)

    def load_image(
        self,
        png_path: Union[str, Path],
    ) -> pyvips.Image:
        '''
        Load and verify an image using pyvips
        '''
        try:
            image = Vips.Image.pngload(str(png_path))
        except pyvips.error.Error as pyvips_error:
            raise ComposingException(
                f'Cannot load {png_path}: {pyvips_error.message}') from None
        except UnicodeDecodeError:
            raise ComposingException(
                f'Cannot load {png_path} with UnicodeDecodeError, '
                'please report your setup at '
                'https://github.com/libvips/pyvips/issues/80') from None
        if image.interpretation != 'srgb':
            image = image.colourspace('srgb')

        try:
            if not image.hasalpha():
                image = image.addalpha()
            if image.get_typeof('icc-profile-data') != 0:
                image = image.icc_transform('srgb')
        except Vips.Error as vips_error:
            log.error(
                '%s: %s',
                png_path, vips_error)

        if (image.width != self.sprite_width or
                image.height != self.sprite_height):
            log.error(
                '%s is %sx%s, but %s sheet sprites have to be %sx%s.',
                png_path, image.width, image.height, self.name,
                self.sprite_width, self.sprite_height)

        return image

    def process_json(
        self,
        filepath: Path,
    ) -> None:
        '''
        Load and store tile entries from the file
        '''
        with open(filepath, 'r', encoding="utf-8") as file:
            try:
                tile_entries = json.load(file)
            except Exception:
                log.error(
                    'error loading %s',
                    filepath)
                raise

            if not isinstance(tile_entries, list):
                tile_entries = [tile_entries]
            for input_entry in tile_entries:
                self.tile_entries.append(
                    TileEntry(self, input_entry, filepath))

    def write_composite_png(self) -> bool:
        '''
        Compose and save tilesheet PNG if there are sprites to work with
        '''
        if not self.sprites:
            return False

        # count empty spaces in the last row
        self.tileset.pngnum += self.sprites_across - \
            ((len(self.sprites) % self.sprites_across) or self.sprites_across)

        if self.tileset.only_json:
            return True

        sheet_image = Vips.Image.arrayjoin(
            self.sprites, across=self.sprites_across)

        pngsave_args = PNGSAVE_ARGS.copy()

        if self.tileset.palette:
            pngsave_args['palette'] = True

        sheet_image.pngsave(str(self.output), **pngsave_args)

        if self.tileset.palette_copies and not self.tileset.palette:
            sheet_image.pngsave(
                str(self.output) + '8', palette=True, **pngsave_args)

        return True


class TileEntry:
    '''
    Tile entry handling
    '''
    def __init__(
        self,
        tilesheet: Tilesheet,
        data: dict,
        filepath: Union[str, Path],
    ) -> None:
        self.tilesheet = tilesheet
        self.data = data
        self.filepath = filepath

    def convert(
        self,
        entry: Union[dict, None] = None,
        prefix: str = '',
    ) -> Optional[dict]:
        '''
        Recursively compile input into game-compatible objects in-place
        '''
        if entry is None:
            entry = self.data

        entry_ids = entry.get('id')
        fg_layer = entry.get('fg')
        bg_layer = entry.get('bg')

        if not entry_ids or (not fg_layer and not bg_layer):
            log.warning(
                'skipping empty entry in %s%s',
                self.filepath,
                f' with IDs {prefix}{entry_ids} ' if entry_ids else '',
            )
            return None

        # make sure entry_ids is a list
        if entry_ids:
            if not isinstance(entry_ids, list):
                entry_ids = [entry_ids]

        # convert fg value
        if fg_layer:
            entry['fg'] = list_or_first(self.convert_entry_layer(fg_layer))
        else:
            # don't pop at the start because that affects order of the keys
            entry.pop('fg', None)

        # convert bg value
        if bg_layer:
            entry['bg'] = list_or_first(self.convert_entry_layer(bg_layer))
        else:
            # don't pop at the start because that affects order of the keys
            entry.pop('bg', None)

        # recursively convert additional_tiles value
        additional_entries = entry.get('additional_tiles', [])
        for additional_entry in additional_entries:
            # recursive part
            self.convert(additional_entry, f'{entry_ids[0]}_')

        # remember processed IDs and remove duplicates
        for entry_id in entry_ids:
            full_id = f'{prefix}{entry_id}'

            if full_id not in self.tilesheet.tileset.processed_ids:
                self.tilesheet.tileset.processed_ids.append(full_id)

            else:
                entry_ids.remove(entry_id)

                if self.tilesheet.is_filler:
                    if self.tilesheet.tileset.obsolete_fillers:
                        log.warning(
                            'skipping filler for %s from %s',
                            full_id, self.filepath)

                else:
                    log.error(
                        '%s encountered more than once, last time in %s',
                        full_id, self.filepath)

        # return converted entry if there are new IDs
        if entry_ids:
            entry['id'] = list_or_first(entry_ids)
            return entry

        return None

    def convert_entry_layer(
        self,
        entry_layer: Union[list, str],
    ) -> list:
        '''
        Convert sprite names to sprite indexes in one fg or bg tile entry part
        '''
        output = []

        if isinstance(entry_layer, list):
            # "fg": [ "f_fridge_S", "f_fridge_W", "f_fridge_N", "f_fridge_E" ]
            for layer_part in entry_layer:
                if isinstance(layer_part, dict):
                    # weighted random variations
                    variations, valid = self.convert_random_variations(
                        layer_part.get('sprite'))
                    if valid:
                        layer_part['sprite'] = \
                            variations[0] if len(variations) == 1 \
                            else variations
                        output.append(layer_part)
                else:
                    self.append_sprite_index(layer_part, output)
        else:
            # "bg": "t_grass"
            self.append_sprite_index(entry_layer, output)

        return output

    def convert_random_variations(
        self,
        sprite_names: Union[list, str],
    ) -> Tuple[list, bool]:
        '''
        Convert list of random weighted variation objects
        '''
        valid = False
        converted_variations = []

        if isinstance(sprite_names, list):
            # list of rotations
            converted_variations = []
            for sprite_name in sprite_names:
                valid |= self.append_sprite_index(
                    sprite_name, converted_variations)
        else:
            # single sprite
            valid = self.append_sprite_index(
                sprite_names, converted_variations)
        return converted_variations, valid

    def append_sprite_index(
        self,
        sprite_name: str,
        entry: list,
    ) -> bool:
        '''
        Get sprite index by sprite name and append it to entry
        '''
        if sprite_name:
            sprite_index = self.tilesheet.tileset\
                .pngname_to_pngnum.get(sprite_name, 0)
            if sprite_index:
                sheet_type = 'filler' if self.tilesheet.is_filler else 'main'
                try:
                    self.tilesheet.tileset\
                        .unreferenced_pngnames[sheet_type].remove(sprite_name)
                except ValueError:
                    pass

                entry.append(sprite_index)
                return True

            log.error(
                '%(name)s.png file for %(name)s value from %(path)s '
                'was not found. It will not be added to %(target)s',
                {
                    'name': sprite_name,
                    'path': self.filepath,
                    'target': self.tilesheet.tileset.output_conf_file,
                })

        return False


def main() -> Union[int, ComposingException]:
    """
    Called when the script is executed directly
    """
    # read arguments and initialize objects
    arg_parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    arg_parser.add_argument(
        'source_dir', type=Path,
        help='Tileset source files directory path')
    arg_parser.add_argument(
        'output_dir', nargs='?', type=Path,
        help='Output directory path')
    arg_parser.add_argument(
        '--use-all', dest='use_all', action='store_true',
        help='Add unused images with id being their basename')
    arg_parser.add_argument(
        '--obsolete-fillers', dest='obsolete_fillers', action='store_true',
        help='Warn about obsoleted fillers')
    arg_parser.add_argument(
        '--palette-copies', dest='palette_copies', action='store_true',
        help='Produce copies of tilesheets quantized to 8bpp colormaps.')
    arg_parser.add_argument(
        '--palette', dest='palette', action='store_true',
        help='Quantize all tilesheets to 8bpp colormaps.')
    arg_parser.add_argument(
        '--format-json', dest='format_json', action='store_true',
        help='Use either CDDA formatter or Python json.tool '
        'to format the tile_config.json')
    arg_parser.add_argument(
        '--only-json', dest='only_json', action='store_true',
        help='Only output the tile_config.json')
    arg_parser.add_argument(
        '--fail-fast', dest='fail_fast', action='store_true',
        help='Stop immediately after an error has occurred')
    arg_parser.add_argument(
        '--loglevel', dest='loglevel',
        choices=['INFO', 'WARNING', 'ERROR'],  # 'DEBUG', 'CRITICAL'
        default='WARNING',
        help="set verbosity level")
    arg_parser.add_argument(
        "--feedback", dest="feedback",
        choices=['SILENT', 'CONCISE', 'VERBOSE'],
        default="SILENT",
        help="When SILENT no output to terminal is given (run silently)."
        " CONCISE displays limited progress feedbeck with no dependency"
        " required. VERBOSE displays progress bar(s) that require TQDM"
        " module (will prompt for installation if absent).")

    args_dict = vars(arg_parser.parse_args())

    dictConfig(LOGGING_CONFIG)
    log.setLevel(getattr(logging, args_dict.get('loglevel')))
    log_tracker = LevelTrackingFilter()
    log.addFilter(log_tracker)

    if args_dict.get('fail_fast'):
        failfast_handler = FailFastHandler()
        failfast_handler.setLevel(logging.ERROR)
        log.addHandler(failfast_handler)

    if args_dict['feedback'] == 'SILENT':
        global run_silent
        run_silent = True

    if args_dict['feedback'] == 'VERBOSE':
        run_silent = False
        args_dict['feedback'] = setup_progress_bar()  # may fallback to CONCISE

    if args_dict['feedback'] == 'CONCISE':
        run_silent = False
        global no_tqdm
        no_tqdm = True  # equal to concise display

    # compose the tileset
    try:
        tileset_worker = Tileset(
            source_dir=Path(args_dict.get('source_dir')),
            output_dir=Path(
                args_dict.get('output_dir') or
                args_dict.get('source_dir')
            ),
            use_all=args_dict.get('use_all', False),
            obsolete_fillers=args_dict.get('obsolete_fillers', False),
            palette_copies=args_dict.get('palette_copies', False),
            palette=args_dict.get('palette', False),
            format_json=args_dict.get('format_json', False),
            only_json=args_dict.get('only_json', False),
        )
        tileset_worker.compose()
    except ComposingException as exception:
        return exception

    if log_tracker.level >= logging.ERROR:
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
"""Print a Markdown or CSV table of JSON values from the given keys.
Run this script with -h for full usage information.

Examples with basic field names:

    %(prog)s --type=TOOL id volume weight longest_side
    %(prog)s --type=ARMOR --format=csv id encumbrance coverage
    %(prog)s --type=COMESTIBLE id fun calories quench healthy

Examples with nested attributes:

    %(prog)s --type=ARMOR name.str pocket_data.moves
    %(prog)s --type=MAGAZINE name.str pocket_data.ammo_restriction

"""

import argparse
import csv
import sys
import util


# Command-line arguments
parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument(
    "columns", metavar="column_key", nargs="+",
    help="list of JSON object keys to be columns in the table")
parser.add_argument(
    "--fnmatch",
    default="*.json",
    help="override with glob expression to select a smaller fileset")
parser.add_argument(
    "-f", "--format",
    default="md",
    help="output format: 'md' for markdown, 'csv' for comma-separated")
parser.add_argument(
    "-t", "--type",
    help="only include JSON data matching these types, separated by comma",
    type=lambda s: list([i for i in s.split(',')]))
parser.add_argument(
    "--nonestring",
    default="None",
    help="what to output when value is None")
parser.add_argument(
    "--noheader",
    dest='with_header', action='store_false',
    help="do not output table header")
parser.add_argument(
    "--tileset",
    dest='tileset_types_only', action='store_true',
    help="override --type filter with a set of types required for a tileset")
parser.set_defaults(with_header=True, tileset_types_only=False)


I18N_DICT_KEYS = ('str', 'str_sp', 'str_pl', 'ctxt', '//~')
I18N_DICT_KEYS_SET = set(I18N_DICT_KEYS)
TILESET_TYPES = [
    "AMMO", "ARMOR", "BATTERY", "BIONIC_ITEM", "bionic", "BOOK", "COMESTIBLE",
    "ENGINE", "field_type", "furniture", "gate", "GENERIC", "GUN", "GUNMOD",
    "MAGAZINE", "MONSTER", "mutation", "PET_ARMOR", "SPELL", "terrain", "TOOL",
    "TOOL_ARMOR", "TOOLMOD", "trap", "vehicle_part", "WHEEL"]


def item_values(item, fields, none_string="None"):
    """Return item values from within the given fields, converted to strings.

    Fields may be plain string or numeric values:

        >>> item_values({'name': 'sword', 'length_cm': 90},
        ...             ['name', 'length_cm'])
        ['sword', '90']

    Fields may also be nested objects; subkeys may be referenced like
    key.subkey:

        >>> item_values({'loc': {'x': 5, 'y': 10}}, ['loc.x', 'loc.y'])
        ['5', '10']

    Fields with a nested object list have their values combined with "/":

        >>> item_values({'locs': [{'x': 4, 'y': 8}, {'x': 5, 'y': 10}] },
        ...             ['locs.x', 'locs.y'])
        ['4 / 5', '8 / 10']

    Fields may include both plain and dotted keys:

        >>> item_values({'id': 'd6', 'name': {'str': 'die', 'str_pl': 'dice'}},
        ...             ['id', 'name.str', 'name.str_pl'])
        ['d6', 'die', 'dice']

    """
    values = []
    for field in fields:
        if "." in field:
            subkeys = field.split(".")
        else:
            subkeys = [field]
        # Descend into dotted keys
        it = item
        for subkey in subkeys:
            # Is it a dict with this subkey?
            if isinstance(it, dict) and subkey in it:
                it = it[subkey]
            # Is it a list of dicts having this subkey?
            elif isinstance(it, list) and all(subkey in o for o in it):
                # Pull from all subkeys, or just the one
                if len(it) == 1:
                    if isinstance(it[0], dict):
                        it = it[0][subkey]
                    else:
                        it = it[0]
                else:
                    it = [i[subkey] for i in it]
            # Stop if any subkey is not found
            else:
                it = none_string
                break

        if isinstance(it, dict):
            if set(it.keys()) <= I18N_DICT_KEYS_SET:
                # it dict contains only i18zed values
                first_good_value = None
                for k in I18N_DICT_KEYS:
                    value = it.get(k, None)
                    if value:
                        first_good_value = value
                        break
                values.append("%s" % first_good_value or none_string)
            else:
                # Make dict presentable
                values.append("%s" % it.items())
        # Separate lists with slashes
        elif isinstance(it, list):
            values.append(" / ".join(
                "%s" % i if i is not None else none_string for i in it))
        # Otherwise just force string
        else:
            values.append("%s" % it if it is not None else none_string)

    return values


def get_format_class_by_extension(format_string):
    """

    >>> get_format_class_by_extension('csv')
    <class 'table.CSVFormat'>

    """
    format_name = format_string.upper()
    try:
        format_class = getattr(
            sys.modules[__name__],
            "{}Format".format(format_name))
    except AttributeError:
        sys.exit("Unknown format {}".format(format_name))
    return format_class


class MDFormat:
    """
    Markdown
    | col1 | col2 | col3
    """
    def header(self, columns):
        self.row(columns)
        # Markdown table needs a separator after the header
        print("| --- " * len(columns))

    def row(self, values):
        safe_values = [self.safe_value(v) for v in values]
        print("| " + " | ".join(safe_values))

    def safe_value(self, value):
        """Return value with special characters escaped.

        >>> MDFormat.safe_value(MDFormat, 'x|y|z')
        'x\\\\|y\\\\|z'

        """
        return value.replace('|', '\\|')


class CSVFormat:
    """
    Comma-Separated Values
    col1,col2,"col3,with,commas"
    """
    writer = None

    def __init__(self):
        self.writer = csv.writer(sys.stdout)

    def to_utf8(self, lst):
        return [str(elem) for elem in lst]

    def header(self, columns):
        self.row(columns)

    def row(self, values):
        self.writer.writerow(self.to_utf8(values))


class CDDAValues:
    """Worker class that prints table from provided data"""
    output = None

    def __init__(self, format_string):
        format_class = get_format_class_by_extension(format_string)
        self.output = format_class()

    def print_table(self, data, columns, types_filter,
                    none_string, with_header):
        if with_header:
            self.output.header(columns)
        for item in data:
            if types_filter and item.get('type') not in types_filter:
                continue

            self.output.row(item_values(item, columns, none_string))


if __name__ == "__main__":
    args = parser.parse_args()
    if args.tileset_types_only:
        args.type = TILESET_TYPES

    # Get data (don't care about load errors)
    json_data, _ = util.import_data(json_fmatch=args.fnmatch)

    worker = CDDAValues(args.format)
    worker.print_table(
        json_data, args.columns, args.type, args.nonestring, args.with_header)



from ..write_text import write_text


def parse_clothing_mod(json, origin):
    write_text(json["implement_prompt"], origin,
               comment="Implement clothing modification prompt text")
    write_text(json["destroy_prompt"], origin,
               comment="Destroy clothing modification prompt text")



from ..write_text import write_text, append_comment


def parse_snippet(json, origin):
    text = json["text"]
    if type(text) is not list:
        text = [text]
    comment = json.get("//") or []
    comment_snip = f"Snippet in category \"{json['category']}\""
    comment_name = f"Name of a snippet in category \"{json['category']}\""
    comment_snip = append_comment([comment_snip], comment)
    comment_name = append_comment([comment_name], comment)
    c_format = "schizophrenia" in json
    for snip in text:
        if type(snip) is str:
            write_text(snip, origin, comment=comment_snip, c_format=c_format)
        elif type(snip) is dict:
            if "name" in snip:
                write_text(snip["name"], origin, comment=comment_name,
                           c_format=False)
            write_text(snip["text"], origin, comment=comment_snip,
                       c_format=c_format)



from .enchant import parse_enchant
from ..helper import get_singular_name
from ..write_text import write_text


def parse_bionic(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Name of a bionic")

    if "description" in json:
        write_text(json["description"], origin, c_format=False,
                   comment="Description of bionic \"{}\"".format(name))

    if "enchantments" in json:
        for enchantment in json["enchantments"]:
            parse_enchant(enchantment, origin)



from ..write_text import write_text


def parse_end_screen(json, origin):
    if "added_info" in json:
        for line in json["added_info"]:
            write_text(line[1], origin,
                       comment="String from the end screen, used in ASCII-art,"
                               " so be aware of the string length.")

    if "last_words_label" in json:
        write_text(json["last_words_label"], origin,
                   comment="String used to label the last word input prompt."
                           "ex: \"Last Words:\")")



#!/usr/bin/env python3
"""Translates map coordinate to map file and folder and vice-versa
Run this script with -h for full usage information.

Info about output:
     Map: In game map coordinate number
    File: Where coordinate is found on the maps folder
  Folder: Range of submaps the folder can have
Overmaps: Number of overmaps the folder contains

Examples with map format:

    %(prog)s 6.1.-3.map
    %(prog)s -4.8.1.map
    %(prog)s -9.-9.17.map

Examples using folder format:

    %(prog)s 1.0.0
    %(prog)s -7.3.5
    %(prog)s 8.-9.-1

Examples with map coordinate format:

    %(prog)s "5'107 3'146"
    %(prog)s "7'98 -7'137 5"
    %(prog)s "-1'0 0'4 -8"

Examples of INVALID map coordinate format:

    %(prog)s "3'180 3'17"    -> Number after the ' must be between 0 and 179
    %(prog)s "7'31  5'-17 1" -> Number after the ' must be positive
    %(prog)s 1'13 3'17 1   -> Quotation marks are needed (")
                              to prevent string escaping

Examples with range:

    %(prog)s "0'31 0'0" --range "0'32 0'0"
    %(prog)s "1'5 4'3 5" --range "-7'1 -9'179 -7"
    %(prog)s 7.3.0.map --range -1.73.1.map
"""

import sys
import argparse
import re


def print_info() -> None:
    print(
        """
    The game map is divided in overmaps
    Each overmap is divided in overmap terrain tiles (OTT) in a 180x180 grid
    Each OTT is 24x24 tiles divided into 4 submaps ( terrain )
    Each OTT is a `.map` file
    Each map folder is 32x32 OTTs
    """
    )


def info_map(format: str) -> list:
    mx, cx = [int(n) for n in format.split()[0].split("'")]
    my, cy = [int(n) for n in format.split()[1].split("'")]

    dx = mx * 180 + cx
    dy = my * 180 + cy
    dz = 0

    if len(format.split()) == 3:
        dz = int(format.split()[2])

    return (dx, dy, dz)


def info_file(format: str) -> list:
    cx, cy, cz = [int(n) for n in format.split(".")[:3]]
    return (cx, cy, cz)


def info_folder(format: str) -> None:
    cx, cy, cz = [int(n) for n in format.split(".")]
    min_cx, max_cx = cx * 32, cx * 32 + 31
    min_cy, max_cy = cy * 32, cy * 32 + 31

    o1 = f"{min_cx // 180} {min_cy // 180}"
    o2 = f"{min_cx // 180} {max_cy // 180}"
    o3 = f"{max_cx // 180} {min_cy // 180}"
    o4 = f"{max_cx // 180} {max_cy // 180}"
    ol = [o1]
    if o2 not in ol:
        ol.append(o2)
    if o3 not in ol:
        ol.append(o3)
    if o4 not in ol:
        ol.append(o4)

    print(f"  Folder:  {min_cx}.{min_cy}.{cz}.map -> "
          f"{max_cx}.{max_cy}.{cz}.map")
    print("Overmaps:  ", end="")
    for k, v in enumerate(ol):
        if k == len(ol) - 1:
            print(v)
        else:
            print(v, end=", ")


def get_range_list(value1: int, value2: int) -> list:
    min_value = min(value1, value2)
    max_value = max(value1, value2)

    if min_value == max_value:
        return [min_value]
    else:
        return list(range(min_value, max_value + 1))


def info_range(r1: tuple, r2: tuple) -> None:

    lvx = get_range_list(r1[0], r2[0])
    lvy = get_range_list(r1[1], r2[1])
    lvz = get_range_list(r1[2], r2[2])

    for z in lvz:
        for y in lvy:
            for x in lvx:
                print(f"{x // 32}.{y // 32}.{z}/{x}.{y}.{z}.map", end=" ")
    print("")


if __name__ == "__main__":

    if len(sys.argv) > 1:
        for key, arg in enumerate(sys.argv[1:]):
            if arg[0] == "-" and arg[1].isdigit():
                argvCopy = " " + arg
                sys.argv[key + 1] = argvCopy

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "format",
        type=str,
        nargs="?",
        help="""coordinate format
Accepted formats:
Folders:    x.y.z
Maps:       x.y.z.map
Coordinate: \"x'(0->179) y'(0->179)\" or \"x'(0->179) y'(0->179) z\"
    Default for z is 0
        """,
    )
    parser.add_argument("--info", action="store_true",
                        help="Print info about maps")
    parser.add_argument(
        "--range", help="Print maps in range using maps or coordinate format"
    )
    args = parser.parse_args(sys.argv[1:])

    if not args.format:
        print("Use --help to view usage")
        sys.exit(1)

    args.format = args.format.strip()

    if args.info:
        print_info()

    elif args.range:
        args.range = args.range.strip()
        matchFile = re.compile(r"^-?\d+\.-?\d+\.-?\d+\.map$")
        matchCoord = re.compile(r"^-?\d'-?\d+\s+-?\d+'-?\d+(\s+-?\d+)?$")

        if re.match(matchFile, args.format):
            if re.match(matchFile, args.range):
                retInfoR = info_file(args.range)

            elif re.match(matchCoord, args.range):
                retInfoR = info_map(args.range)

            else:
                print("Invalid format for range, use --help to view usage")
                sys.exit(1)

            retInfoF = info_file(args.format)

        elif re.match(matchCoord, args.format):
            if re.match(matchFile, args.range):
                retInfoR = info_file(args.range)

            elif re.match(matchCoord, args.range):
                retInfoR = info_map(args.range)

            else:
                print("Invalid format for range, use --help  to view usage")
                sys.exit(1)

            retInfoF = info_map(args.format)

        else:
            print("Invalid format, print help to view usage")
            sys.exit(1)

        info_range(retInfoF, retInfoR)

    else:
        args.format = args.format.strip()

        retInfo = False

        # .map file pattern
        if re.match(r"^-?\d+\.-?\d+\.-?\d+\.map$", args.format):
            retInfo = info_file(args.format)

        # .map folder pattern
        elif re.match(r"^-?\d+\.-?\d+\.-?\d+$", args.format):
            info_folder(args.format)

        # map coordinate pattern
        elif re.match(
            r"^-?\d'(\d{0,2}|1[0-7]\d)\s+-?\d+'(\d{0,2}|1[0-7]\d)(\s+-?\d+)?$",
            args.format,
        ):
            retInfo = info_map(args.format)
        elif re.match(r"^-?\d'-?\d+\s+-?\d+'-?\d+(\s+-?\d+)?$", args.format):
            print(
                "Invalid range for map coordination,"
                " a'b where b is positive from 0 to 179"
            )
        else:
            print("Invalid format, print help to view usage")

        if retInfo:
            cx, cy, cz = retInfo

            print(
                f"     Map:  {cx // 180}'{cx - 180 * (cx // 180)}"
                f"  {cy // 180}'{cy - 180 * (cy // 180)}  {cz}"
            )

            print(f"    File:  {cx // 32}.{cy // 32}.{cz}/{cx}.{cy}.{cz}.map")

            info_folder(f"{cx // 32}.{cy // 32}.{cz}")



from ..write_text import write_text


def parse_start_location(json, origin):
    if "name" in json:
        id = json["id"]
        write_text(json["name"], origin,
                   comment="Name of starting location \"{}\"".format(id))



"""Utility stuff for json tools.
"""

import argparse
from collections import Counter, OrderedDict
from fnmatch import fnmatch
import json
import re
import os
from io import StringIO


SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))
JSON_DIR = os.path.normpath(
    os.path.join(SCRIPT_DIR, "..", "..", "data", "json"))
JSON_FNMATCH = "*.json"


def import_data(json_dir=JSON_DIR, json_fmatch=JSON_FNMATCH):
    """Use a UNIX like file match expression to weed out the JSON files.

    returns tuple, first element containing json read, second element
    containing list of any errors found. error list will be empty if
    no errors
    """
    data = []
    errors = []
    candidates = None
    for d_descriptor in os.walk(json_dir):
        d = d_descriptor[0]
        for f in d_descriptor[2]:
            if fnmatch(f, json_fmatch):
                json_file = os.path.join(d, f)
                with open(json_file, "r", encoding="utf-8") as file:
                    try:
                        candidates = json.load(
                            file, object_pairs_hook=OrderedDict)
                    except Exception as err:
                        errors.append(
                            "Problem reading file {},".format(json_file) +
                            " reason: {}".format(err))
                    if type(candidates) is not list:
                        if type(candidates) is OrderedDict:
                            data.append(candidates)
                        else:
                            errors.append(
                                "Problem parsing data from" +
                                " file {},".format(json_file) +
                                " reason: expected a list.")
                    else:
                        data += candidates
    return (data, errors)


def match_primitive_values(item_value, where_value):
    """Perform any odd logic on item matching.
    """
    # Matching interpolation for keyboard constrained input.
    if type(item_value) is str:
        # Direct match
        return bool(re.match(where_value, item_value))
    elif type(item_value) is int or type(item_value) is float:
        # match after string conversion
        return bool(re.match(where_value, str(item_value)))
    elif type(item_value) is bool:
        # help conversion to JSON booleans from the commandline
        return bool(re.match(where_value, str(item_value).lower()))
    else:
        return False


def matches_where(item, where_key, where_value):
    """True if:
    where_key exists AND
    where_value somehow matches (either directly or guesstimated).

    False if:
    no where_key passed in
    where_key not in item
    where_key in item but where_value does not match
    """
    if not where_key:
        return True
    if where_key not in item:
        return False
    # So we have some value.
    item_value = item[where_key]
    # Matching interpolation for keyboard constrained input.
    if type(item_value) is list:
        # 1 level deep.
        for next_level in item_value:
            if match_primitive_values(next_level, where_value):
                return True
        # else...
        return False
    elif type(item_value) is dict:
        # Match against the keys of the dictionary... I question my logic.
        # 1 level deep.
        for next_level in item_value:
            if match_primitive_values(next_level, where_value):
                return True
        # else...
        return False
    else:
        return match_primitive_values(item_value, where_value)


def matches_all_wheres(item, where_fn_list):
    """Takes a list of where functions and attempts to match against them.

    Assumes wheres are the type returned from WhereAction, and the function
    accepts the item to match against.

    True if:
    all where's match (effectively AND)

    False if:
    any where's don't match
    """
    for where_fn in where_fn_list:
        if not where_fn(item):
            return False
    # Must be a match.
    return True


class WhereAction(argparse.Action):
    """An argparse action callback.

    Example application:

    parser.add_argument("where",
        action=WhereAction, nargs='*', type=str, help="where exclusions")
    """

    def where_test_factory(self, where_key, where_value):
        """Wrap the where test we are using and return it as a callable
        function.

        item in the callback is assumed to be what we're testing against.
        """
        def t(item):
            return matches_where(item, where_key, where_value)
        return t

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if not nargs:
            raise ValueError("nargs must be declared")
        super(WhereAction, self).__init__(
            option_strings, dest, nargs=nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        try:
            where_functions = []
            for w in values:
                w_key, w_value = w.split("=")
                where_functions.append(self.where_test_factory(w_key, w_value))
            setattr(namespace, self.dest, where_functions)
        except Exception:
            raise ValueError(
                "Where options are strict. Must be in the form of"
                " 'where_key=where_value'")


def key_counter(data, where_fn_list):
    """Count occurences of keys found in data {list of dicts}
    that also match each where_fn_list {list of fns}.

    Returns a tuple of data.
    """
    stats = Counter()
    matching_data = [i for i in data if matches_all_wheres(i, where_fn_list)]
    for item in matching_data:
        # Check each key for nested data
        for key in item.keys():
            # Skip comments
            if key.startswith('//'):
                continue

            val = item[key]
            # If value is an object, tally key.subkey for all object subkeys
            if type(val) is OrderedDict:
                for subkey in val.keys():
                    if not subkey.startswith('//'):
                        stats[key + '.' + subkey] += 1

            # If value is a list of objects, tally key.subkey for each
            elif type(val) is list:
                if all(type(e) is OrderedDict for e in val):
                    for obj in val:
                        for subkey in obj.keys():
                            if not subkey.startswith('//'):
                                stats[key + '.' + subkey] += 1

            # For anything else, it only counts as one
            else:
                stats[key] += 1

    return stats, len(matching_data)


def item_value_counter(_value):
    """Return a Counter for a single value (int, float, string, dict)
    recursing into nested dicts and lists.
    """
    stats = Counter()
    # String or unicode
    if isinstance(_value, str):
        stats[_value] += 1
    # Cast numbers to strings
    elif type(_value) is int or type(_value) is float:
        stats[str(_value)] += 1
    # Pull all values from objects
    elif type(_value) is OrderedDict:
        stats += list_value_counter(list(_value.values()))
    # Pull values from list of objects or strings
    elif type(_value) is list:
        stats += list_value_counter(_value)
    else:
        raise ValueError("Value '%s' has unknown type %s" %
                         (_value, type(_value)))
    return stats


def list_value_counter(_list):
    """Return a Counter tallying all values in the given {list of dicts}
    or {list of strs}, recursing into nested dicts and lists.
    """
    stats = Counter()
    for elem in _list:
        stats += item_value_counter(elem)
    return stats


def value_counter(data, search_key, where_fn_list):
    """Takes a search_key {str}, and for values found in data {list of dicts}
    that also match each where_fn_list {list of fns} with those keys,
    counts the number of times the value appears.

    Returns a tuple of data.
    """
    stats = Counter()
    matching_data = [i for i in data if matches_all_wheres(i, where_fn_list)]

    # Search key may use one level of dotted notation to descend into an
    # dict or list-of-dict in the matching data

    # Exactly one dot is allowed to look in parent_key.child_key
    if search_key.count('.') == 1:
        parent_key, child_key = search_key.split('.')
    # With no dots, just use the plain key
    elif search_key.count('.') == 0:
        parent_key = search_key
        child_key = None
    else:
        raise argparse.ArgumentError("Only one '.' allowed in search key")

    for item in matching_data:
        if parent_key not in item:
            stats['MISSING'] += 1
            continue

        parent_val = item[parent_key]

        # List of values within this item to tally stats on
        stat_vals = []

        # If this value is a list of objects, pull parent_key.child_key
        # values from all of them to include in stats
        if type(parent_val) is list and all(type(e) is OrderedDict
                                            for e in parent_val):
            for od in parent_val:
                if child_key in od:
                    stat_vals.append(od[child_key])

        # If this value is a single object, get value at parent_key.child_key
        elif type(parent_val) is OrderedDict and child_key in parent_val:
            stat_vals.append(parent_val[child_key])

        # Other kinds of data cannot be indexed by parent_key.child_key
        else:
            stat_vals.append(parent_val)

        # Tally all stats, recursing into nested values
        for val in stat_vals:
            stats += item_value_counter(val)

    return stats, len(matching_data)


def ui_values_to_columns(values, screen_width=80):
    """Take a list of strings and output in fixed width columns.
    """
    max_val_len = len(max(values, key=len)) + 1
    cols = screen_width / max_val_len
    iters = 0
    for v in values:
        print(v.ljust(max_val_len), end=' ')
        iters += 1
        if iters % cols == 0:
            print("")
    print("")


def ui_counts_to_columns(counts):
    """Take a Counter instance and display in single fixed width key:value
    column.
    """
    # Values in left column, counts in right, left
    # column as wide as longest string length.
    key_vals = counts.most_common()
    key_field_len = len(max(list(counts.keys()), key=len)) + 1
    output_template = "%%-%ds: %%s" % key_field_len
    for k_v in key_vals:
        print(output_template % k_v)


class CDDAJSONWriter(object):
    """Essentially a one-off class used to write CDDA formatted JSON output.

    Expects single object, not a list of objects.

    Probable usage:

        print CDDSJSONWriter(some_json).dumps()
    """
    indent = "  "
    indent_multiplier = 0
    buf = None

    def __init__(self, d, indent_multiplier=0):
        self.d = d
        # Should you wish to change the initial indent for whatever reason.
        self.indent_multiplier = indent_multiplier
        # buf is initialized on a call to dumps

    def indented_write(self, s):
        self.buf.write(self.indent * self.indent_multiplier + s)

    def write_key(self, k):
        self.indented_write("\"%s\": " % k)

    def write_primitive_key_val(self, k, v):
        self.write_key(k)
        self.buf.write(json.dumps(v))

    def list_of_lists(self, k, lol):
        self.write_key(k)
        self.buf.write("[\n")
        lol = lol[:]
        while lol:
            self.indent_multiplier += 1
            inner = lol.pop(0)[:]
            self.indented_write("[\n")
            while inner:
                self.indent_multiplier += 1
                item = inner.pop(0)
                # Print each of these on one line
                self.indented_write(json.dumps(item))
                if inner:
                    self.buf.write(",\n")
                self.indent_multiplier -= 1
            self.buf.write("\n")
            self.indented_write("]")
            self.buf.write(",\n" if lol else "\n")
            self.indent_multiplier -= 1
        self.indented_write("]")

    def dumps(self):
        """Format the Cataclysm JSON in as friendly of a JSON way as we can.
        """
        if self.buf:
            self.buf.close()
            self.buf = None

        self.buf = StringIO()
        items = list(self.d.items())
        global indent_multiplier
        self.indented_write("{\n")
        self.indent_multiplier += 1
        while items:
            k, v = items.pop(0)
            # Special cases first.
            if (k == "tools" or k == "components") and type(v) is list:
                self.list_of_lists(k, v)
            else:
                self.write_primitive_key_val(k, v)

            # Trailing comma or not
            if items:
                self.buf.write(",\n")
        self.buf.write("\n")
        self.indent_multiplier -= 1
        self.indented_write("}")
        return self.buf.getvalue()



from ..write_text import write_text


def parse_field_type(json, origin):
    field_names = []

    for fd in json.get("intensity_levels", []):
        if "name" in fd:
            write_text(fd["name"], origin, comment="Field intensity level")
            id_or_name = fd["name"]
        else:
            id_or_name = json["id"]
        field_names.append(id_or_name)
        for eff in fd.get("effects", []):
            if "message" in eff:
                write_text(eff["message"], origin,
                           comment="Player message on effect of field {}"
                           .format(id_or_name))
            if "message_npc" in eff:
                write_text(eff["message_npc"], origin,
                           comment="NPC message on effect of field {}"
                           .format(id_or_name))
    if "npc_complain" in json:
        write_text(json["npc_complain"]["speech"], origin,
                   comment="Field NPC complaint")

    if "bash" in json:
        if "msg_success" in json["bash"]:
            write_text(json["bash"]["msg_success"], origin,
                       comment="Bashing message of fields {}"
                       .format(", ".join(f"\"{n}\"" for n in field_names)))
